# SPDX-License-Identifier: GNU GPL v3

"""
Entry points that allow users to execute GUI or Cli programs
"""

import sys
import time
import logging
from .apps.gui_app import PySideApp
from .apps.cli_app import TerminalApp


logger = logging.getLogger("SGT App")
# FORMAT = '%(asctime)s; %(user)s. %(levelname)s: %(message)s'
FORMAT = '%(asctime)s; %(levelname)s: %(message)s'
DATE_FORMAT = '%Y-%m-%d %H:%M:%S'


def main_gui() -> None:
    """
    Start the graphical user interface application.
    :return:
    """
    # Initialize log collection
    logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s", stream=sys.stdout)
    logging.info("SGT application started running...", extra={'user': 'SGT Logs'})

    # Install CuPy for GPU
    # detect_cuda_and_install_cupy()

    # Start GUI app
    PySideApp.start()

    # Log to show the App stopped
    logging.info("SGT application stopped running.", extra={'user': 'SGT Logs'})


def main_cli() -> None:
    """
    Start the terminal/CMD application.
    :return:
    """
    f_name = str('sgt_app' + str(time.time()).replace('.', '', 1) + '.log')
    logging.basicConfig(filename=f_name, encoding='utf-8', level=logging.INFO, format=FORMAT, datefmt=DATE_FORMAT)
    logging.info("SGT application started running...", extra={'user': 'SGT Logs'})

    TerminalApp.start()
    logging.info("SGT application stopped running.", extra={'user': 'SGT Logs'})


# SPDX-License-Identifier: GNU GPL v3

"""
Terminal interface implementations
"""

import sys
import logging
from optparse import OptionParser

from .base_worker import BaseWorker
from .base_contoller import BaseController
from ..compute.graph_analyzer import GraphAnalyzer
from ..imaging.image_processor import ImageProcessor
from ..networks.fiber_network import FiberNetworkBuilder
from ..utils.config_loader import strict_read_config_file

logger = logging.getLogger("SGT App")
#logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s", stream=sys.stdout)

class TerminalApp(BaseController):
    """Exposes the terminal interface for StructuralGT."""

    def __init__(self, config_path: str):
        """
        Exposes methods for running StructuralGT tasks
        :param config_path: the path to the configuration file
        """
        super().__init__(config_file=config_path)
        self._task_worker = BaseWorker()
        self._task_worker.inProgressSignal.connect(TerminalApp.update_progress)
        self._task_worker.taskFinishedSignal.connect(TerminalApp.task_finished)
        self._selected_sgt_obj_index = 0
        # self.showAlertSignal.connect(self.show_alert)

    @property
    def task_worker(self):
        return self._task_worker

    def check_image_files(self, img_path: str="", img_dir: str = "", out_dir: str = ""):
        """"""
        # 1. Verify config file
        config_file_ok = strict_read_config_file(self._config_file, self.update_progress)
        if not config_file_ok:
            sys.exit('Usage: StructuralGT-cli -f datasets/InVitroBioFilm.png -c sgt_configs.ini -t 2 -o results/')

        # 2. Get images and process them
        if img_path != "":
            self.add_single_image(img_path, out_dir)
        elif img_dir != "":
            self.add_multiple_images(img_dir, out_dir)
        else:
            self.update_progress(-1, "No image(s) found in the path/image folder provided!")
            sys.exit('System exit')

    @staticmethod
    def task_finished(success_val: bool, result: None|list|ImageProcessor|FiberNetworkBuilder|GraphAnalyzer) -> None:
        """
        Handler function for sending updates/signals on termination of tasks.
        Args:
            success_val:
            result:

        Returns:

        """
        if not success_val:
            if type(result) is list:
                logging.info(result[0] + ": " + result[1], extra={'user': 'SGT Logs'})
                TerminalApp.update_progress(101, f"{result[0]}: {result[1]}")
        else:
            if type(result) is ImageProcessor:
                TerminalApp.update_progress(100, "Graph extracted successfully!")
            elif type(result) is GraphAnalyzer:
                TerminalApp.update_progress(100, "GT PDF successfully generated! Check out generated PDF in 'Output Dir'.")
            elif type(result) is dict:
                TerminalApp.update_progress(100, "All GT PDF successfully generated! Check out generated PDF in 'Output Dir'.")
            elif type(result) is list:
                # Histogram data
                pass
            else:
                pass

    @staticmethod
    def update_progress(progress_val: int, msg: str) -> None:
        """
        Simple method to display progress updates.

        Args:
            progress_val (int): progress value
            msg (str): progress message
        Returns:
             None:
        """
        if 0 <= progress_val <= 100:
            print(f"{progress_val} %: {msg}")
            logging.info(f"{progress_val} %: {msg}", extra={'user': 'SGT Logs'})
        elif progress_val > 100:
            print(f"{msg}")
            logging.info(f"{msg}", extra={'user': 'SGT Logs'})
        else:
            print(f"Error: {msg}")
            logging.exception(f"{msg}", extra={'user': 'SGT Logs'})

    @classmethod
    def start(cls) -> None:
        """Initializes and starts the terminal/CMD the StructuralGT application."""

        # Retrieve user settings
        opt_parser = OptionParser()
        opt_parser.add_option('-f', '--inputFile',
                             dest='file_path',
                             help='path to image file',
                             default="../datasets/InVitroBioFilm.png",
                             type='string')
        opt_parser.add_option('-d', '--inputDir',
                             dest='img_dir_path',
                             help='path to folder containing images',
                             default="",
                             type='string')
        opt_parser.add_option('-o', '--outputDir',
                              dest='output_dir',
                              help='path to folder for saving output files. If not provided, output files will be saved in input dir.',
                              default="",
                              type='string')
        opt_parser.add_option('-s', '--allowAutoScale',
                             dest='auto_scale',
                             help='allow automatic scaling of images',
                             default=1,
                             type='int')
        opt_parser.add_option('-t', '--runTask',
                              dest='run_task',
                              help='you can run the following tasks: (1) extract graph; (2) compute GT metrics.',
                              default=2,
                              type='int')
        opt_parser.add_option('-c', '--config',
                              dest='config_file',
                              help='path to config file',
                              default="",
                              type='string')
        # opt_parser.add_option('-m', '--runMultiGT',
        #                      dest='run_multi_gt',
        #                      help='run compute GT parameters on multiple images',
        #                      default=0,
        #                      type='int')
        # opt_parser.add_option('-i', '--selectedImgIndex',
        #                      dest='sel_img_idx',
        #                      help='index of selected image',
        #                      default=0,
        #                      type='int')
        (cfg, args) = opt_parser.parse_args()
        cfg.auto_scale = bool(cfg.auto_scale)
        # cfg.run_multi_gt = bool(cfg.run_multi_gt)

        # 1. Create Terminal App
        term_app = cls(cfg.config_file)

        # 2. Verify image files
        term_app.check_image_files(img_path=cfg.file_path, img_dir=cfg.img_dir_path, out_dir=cfg.output_dir)

        # 3. Execute specific task
        if cfg.run_task == 0:
            pass
        elif cfg.run_task == 1:
            sgt_obj = term_app.get_selected_sgt_obj()
            term_app._task_worker.task_extract_graph(sgt_obj.ntwk_p)
        elif cfg.run_task == 2:
            run_multi_gt = True if cfg.img_dir_path != "" else False
            if run_multi_gt:
                term_app.replicate_sgt_configs()
                term_app._task_worker.task_compute_multi_gt(term_app._sgt_objs)
            else:
                term_app._task_worker.task_compute_gt(term_app.get_selected_sgt_obj())
        else:
            term_app.update_progress(-1, "Invalid GT task selected! System will exit.")
            sys.exit('System exit')



from PIL import Image, ImageQt  # Import ImageQt for conversion
from PySide6.QtGui import QPixmap
from PySide6.QtQuick import QQuickImageProvider


class ImageProvider(QQuickImageProvider):

    def __init__(self, img_controller):
        super().__init__(QQuickImageProvider.ImageType.Pixmap)
        self._pixmap = QPixmap()
        self._img_controller = img_controller
        self._img_controller.changeImageSignal.connect(self.handle_change_image)

    def handle_change_image(self):
        if len(self._img_controller.sgt_objs) > 0:
            img_cv = None
            sgt_obj = self._img_controller.get_selected_sgt_obj()
            ntwk_p = sgt_obj.ntwk_p
            sel_img_batch = ntwk_p.selected_batch
            if ntwk_p.selected_batch_view == "original":
                # Calculate image histogram in different thread
                self._img_controller.compute_img_histogram()
                images = ntwk_p.image_3d
                if self._img_controller.is_img_3d():
                    self._img_controller.img3dGridModel.reset_data(images, sel_img_batch.selected_images_idx)
                else:
                    # 2D, Do not use if 3D
                    img_cv = images[0] if len(images) > 0 else None
            elif ntwk_p.selected_batch_view == "binary":
                # Apply filters
                ntwk_p.apply_img_filters(filter_type=2)
                # Calculate image histogram in different thread
                self._img_controller.compute_img_histogram()
                bin_images = ntwk_p.binary_image_3d
                if self._img_controller.is_img_3d():
                    self._img_controller.img3dGridModel.reset_data(bin_images, sel_img_batch.selected_images_idx)
                else:
                    # 2D, Do not use if 3D
                    img_cv = bin_images[0] if len(bin_images) > 0 else None
            elif ntwk_p.selected_batch_view == "processed":
                # Apply filters
                ntwk_p.apply_img_filters(filter_type=1)
                # Calculate image histogram in different thread
                self._img_controller.compute_img_histogram()
                mod_images = ntwk_p.processed_image_3d
                if self._img_controller.is_img_3d():
                    self._img_controller.img3dGridModel.reset_data(mod_images, sel_img_batch.selected_images_idx)
                else:
                    # 2D, Do not use if 3D
                    img_cv = mod_images[0] if len(mod_images) > 0 else None
            elif ntwk_p.selected_batch_view == "graph":
                # If any is None, start the task
                self._img_controller.showImageHistogramSignal.emit(False)
                if ntwk_p.graph_obj.img_ntwk is None:
                    self._img_controller.run_extract_graph()
                    # Wait for the task to finish
                    return
                else:
                    net_images = [ntwk_p.graph_obj.img_ntwk]
                    self._img_controller.img3dGridModel.reset_data(net_images, sel_img_batch.selected_images_idx)
                    img_cv = net_images[0]
            else:
                return

            if img_cv is not None:
                # Create Pixmap image
                img = Image.fromarray(img_cv)
                self._pixmap = ImageQt.toqpixmap(img)

            # Acknowledge the image load and send the signal to update QML
            self._img_controller._img_loaded = True
            self._img_controller.imageChangedSignal.emit()
        else:
            self._img_controller._img_loaded = False
        self._img_controller._applying_changes = False

    def requestPixmap(self, img_id, requested_size, size):
        return self._pixmap



# SPDX-License-Identifier: GNU GPL v3

"""
Loads default configurations from 'configs.ini' file
"""

import os
import configparser
from typing import Union

from .sgt_utils import verify_path


def strict_read_config_file(config_path, update_func=None) -> bool:
    """
    Strictly read the contents of the 'configs.ini' file, otherwise stop execution.

    Args:
        config_path (str): path to config file
        update_func (function): function that will be called to give message updates

    Returns:
        ConfigParser object or None if an error occurs.
    """
    if config_path == "":
        return True

    success, result = verify_path(config_path)
    if not success:
        if update_func is not None:
            update_func(-1, f"File Error: unable to find config file {config_path}.")
        return False

    config = configparser.ConfigParser()
    config_file = result
    try:
        config.read(config_file)
        return True
    except configparser.Error:
        if update_func is not None:
            update_func(-1, f"Unable to read the configs from {config_file}.")
        return False


def read_config_file(config_path):
    """Read the contents of the 'configs.ini' file"""
    config = configparser.ConfigParser()
    success, result = verify_path(config_path)
    if success:
        config_file = result
    else:
        # print(f"File Error: unable to find config file {config_path}. Using the default config file")
        # Get the directory of the current script
        script_dir = os.path.dirname(os.path.abspath(__file__))
        config_path = 'configs.ini'
        config_file = os.path.join(script_dir, config_path)
    # Load the default configuration from the file
    try:
        config.read(config_file)
        return config
    except configparser.Error:
        # print(f"Unable to read the configs from {config_file}.")
        return None


def load_img_configs(cfg_path: str = ""):
    """Image Detection settings"""

    options_img: dict[str, dict[str, Union[int, float]]] = {
        "threshold_type": {"id": "threshold_type", "type": "binary-filter", "text": "", "visible": 1, "value": 0 },
        "global_threshold_value": {"id": "global_threshold_value", "type": "binary-filter", "text": "", "visible": 1, "value": 128 },
        "adaptive_local_threshold_value": {"id": "adaptive_local_threshold_value", "type": "binary-filter", "text": "", "visible": 1, "value": 11 },
        "otsu": {"id": "otsu", "type": "binary-filter", "text": "", "visible": 0, "value": 0},
        "apply_dark_foreground": {"id": "apply_dark_foreground", "type": "binary-filter", "text": "", "visible": 1, "value": 0},

        "apply_autolevel": {"id": "apply_autolevel", "type": "image-filter", "text": "Autolevel", "value": 0,
                            "dataId": "autolevel_blurring_size", "dataValue": 3, "minValue": 1, "maxValue": 7, "stepSize": 2},
        "apply_gaussian_blur": {"id": "apply_gaussian_blur", "type": "image-filter", "text": "Gaussian", "value": 0,
                                "dataId": "gaussian_blurring_size", "dataValue": 3, "minValue": 1, "maxValue": 7, "stepSize": 2 },
        "apply_laplacian_gradient": {"id": "apply_laplacian_gradient", "type": "image-filter", "text": "Laplacian",
                                     "value": 0, "dataId": "laplacian_kernel_size", "dataValue": 3, "minValue": 1, "maxValue": 7,  "stepSize": 2 },
        "apply_lowpass_filter": {"id": "apply_lowpass_filter", "type": "image-filter", "text": "Lowpass", "value": 0,
                                 "dataId": "lowpass_window_size", "dataValue": 10, "minValue": 0, "maxValue": 1000,  "stepSize": 1 },
        "apply_gamma": {"id": "apply_gamma", "type": "image-filter", "text": "LUT Gamma", "value": 1, "dataId": "adjust_gamma",
                        "dataValue": 1.0, "minValue": 0.01, "maxValue": 5.0, "stepSize": 0.01  },
        "apply_sobel_gradient": {"id": "apply_sobel_gradient", "type": "image-filter", "text": "Sobel", "value": 0,
                                 "dataId": "sobel_kernel_size", "dataValue": 3, "minValue": 1, "maxValue": 7,  "stepSize": 2 },
        "apply_median_filter": {"id": "apply_median_filter", "type": "image-filter", "text": "Median", "value": 0 },
        "apply_scharr_gradient": {"id": "apply_scharr_gradient", "type": "image-filter", "text": "Scharr", "value": 0},

        "brightness_level": {"id": "brightness_level", "type": "image-control", "text": "Brightness", "value": 0 },
        "contrast_level": {"id": "contrast_level", "type": "image-control", "text": "Contrast", "value": 0 },

        "scale_value_nanometers": {"id": "scale_value_nanometers", "type": "image-property", "text": "Scalebar (nm)", "visible": 1, "value": 0.0 },
        "scalebar_pixel_count": {"id": "scalebar_pixel_count", "type": "image-property", "text": "Scalebar Pixel Count", "visible": 1, "value": 1 },
        "resistivity": {"id": "resistivity", "type": "image-property", "text": "Resistivity (<html>&Omega;</html>m)", "visible": 1, "value": 1.0 },
        "pixel_width": {"id": "pixel_width", "type": "image-property", "text": "", "visible": 0, "value": 1.0},  # * (10**-9)  # 1 nanometer

        "save_images": {"id": "save_images", "type": "file-options", "text": "Save All Images", "visible": 1, "value": 0},
    }

    # Load configuration from the file
    config = read_config_file(cfg_path)
    if config is None:
        return options_img

    try:
        options_img["threshold_type"]["value"] = int(config.get('filter-settings', 'threshold_type'))
        options_img["global_threshold_value"]["value"] = int(config.get('filter-settings', 'global_threshold_value'))
        options_img["adaptive_local_threshold_value"]["value"] = int(config.get('filter-settings', 'adaptive_local_threshold_value'))
        options_img["apply_dark_foreground"]["value"] = int(config.get('filter-settings', 'apply_dark_foreground'))

        options_img["apply_gamma"]["value"] = int(config.get('filter-settings', 'apply_gamma'))
        options_img["apply_gamma"]["dataValue"] = float(config.get('filter-settings', 'adjust_gamma'))
        options_img["apply_autolevel"]["value"] = int(config.get('filter-settings', 'apply_autolevel'))
        options_img["apply_autolevel"]["dataValue"] = int(config.get('filter-settings', 'blurring_window_size'))
        options_img["apply_laplacian_gradient"]["value"] = int(config.get('filter-settings', 'apply_laplacian_gradient'))
        options_img["apply_laplacian_gradient"]["dataValue"] = 3
        options_img["apply_sobel_gradient"]["value"] = int(config.get('filter-settings', 'apply_sobel_gradient'))
        options_img["apply_sobel_gradient"]["dataValue"] = 3
        options_img["apply_gaussian_blur"]["value"] = int(config.get('filter-settings', 'apply_gaussian_blur'))
        options_img["apply_gaussian_blur"]["dataValue"] = int(config.get('filter-settings', 'blurring_window_size'))
        options_img["apply_lowpass_filter"]["value"] = int(config.get('filter-settings', 'apply_lowpass_filter'))
        options_img["apply_lowpass_filter"]["dataValue"] = int(config.get('filter-settings', 'filter_window_size'))

        options_img["apply_scharr_gradient"]["value"] = int(config.get('filter-settings', 'apply_scharr_gradient'))
        options_img["apply_median_filter"]["value"] = int(config.get('filter-settings', 'apply_median_filter'))

        options_img["brightness_level"]["value"] = int(config.get('filter-settings', 'brightness_level'))
        options_img["contrast_level"]["value"] = int(config.get('filter-settings', 'contrast_level'))
        options_img["scale_value_nanometers"]["value"] = float(config.get('filter-settings', 'scale_value_nanometers'))
        options_img["scalebar_pixel_count"]["value"] = int(config.get('filter-settings', 'scalebar_pixel_count'))
        options_img["resistivity"]["value"] = float(config.get('filter-settings', 'resistivity'))

        options_img["save_images"]["value"] = int(config.get('file-options', 'save_images'))

        return options_img
    except configparser.NoSectionError:
        return options_img


def load_gte_configs(cfg_path: str = ""):
    """Graph Extraction Settings"""

    options_gte = {
        "has_weights": {"id": "has_weights", "type": "graph-extraction", "text": "Add Weights", "value": 0,
                        "items": [
                            {"id": "DIA", "text": "by diameter", "value": 1},
                            {"id": "AREA", "text": "by area", "value": 0},
                            {"id": "LEN", "text": "by length", "value": 0},
                            {"id": "ANGLE", "text": "by angle", "value": 0},
                            {"id": "INV-LEN", "text": "by inverse-length", "value": 0},
                            {"id": "FIX-CON", "text": "by conductance", "value": 0},
                            {"id": "RES", "text": "by resistance", "value": 0},
                        ]},
        "merge_nearby_nodes": {"id": "merge_nearby_nodes", "type": "graph-extraction", "text": "Merge Nearby Nodes", "value": 1, "items": [{"id": "merge_node_radius_size", "text": "", "value": 2}]},
        "prune_dangling_edges": {"id": "prune_dangling_edges", "type": "graph-extraction", "text": "Prune Dangling Edges", "value": 1, "items": [{"id": "prune_max_iteration_count", "text": "", "value": 500}]},
        "remove_disconnected_segments": {"id": "remove_disconnected_segments", "type": "graph-extraction", "text": "Remove Disconn. Segments", "value": 1, "items": [{"id": "remove_object_size", "text": "", "value": 500}]},
        "remove_self_loops": {"id": "remove_self_loops", "type": "graph-extraction", "text": "Remove Self Loops", "value": 1},
        "display_node_id": {"id": "display_node_id", "type": "graph-extraction", "text": "Display Node ID", "value": 0},
        "add_width_thickness": {"id": "add_width_thickness", "type": "graph-extraction", "text": "Show Width Thickness", "value": 0},

        "export_edge_list": {"id": "export_edge_list", "type": "file-options", "text": "Export Edge List", "value": 0},
        "export_node_positions": {"id": "export_node_positions", "type": "file-options", "text": "Export Node Positions", "value": 0},
        "export_as_gexf": {"id": "export_as_gexf", "type": "file-options", "text": "Export as gexf", "value": 0},
        "export_adj_mat": {"id": "export_adj_mat", "type": "file-options", "text": "Export Adjacency Matrix", "value": 0},
        "export_as_gsd": {"id": "export_as_gsd", "type": "file-options", "text": "Export as GSD/HOOMD", "value": 0},
    }

    # Load configuration from the file
    config = read_config_file(cfg_path)
    if config is None:
        return options_gte

    try:

        options_gte["merge_nearby_nodes"]["value"] = int(config.get('extraction-settings', 'merge_nearby_nodes'))
        options_gte["merge_nearby_nodes"]["items"][0]["value"] = int(
            config.get('extraction-settings', 'merge_node_radius_size'))
        options_gte["prune_dangling_edges"]["value"] = int(config.get('extraction-settings', 'prune_dangling_edges'))
        options_gte["prune_dangling_edges"]["items"][0]["value"] = int(
            config.get('extraction-settings', 'prune_max_iteration_count'))
        options_gte["remove_disconnected_segments"]["value"] = int(
            config.get('extraction-settings', 'remove_disconnected_segments'))
        options_gte["remove_disconnected_segments"]["items"][0]["value"] = int(config.get('extraction-settings', 'remove_object_size'))
        options_gte["remove_self_loops"]["value"] = int(config.get('extraction-settings', 'remove_self_loops'))
        options_gte["has_weights"]["value"] = int(config.get('extraction-settings', 'add_weights'))
        weight_type = str(config.get('extraction-settings', 'weight_type'))
        for i in range(len(options_gte["has_weights"]["items"])):
            options_gte["has_weights"]["items"][i]["value"] = 1 if options_gte["has_weights"]["items"][i]["id"] == weight_type else 0
        options_gte["display_node_id"]["value"] = int(config.get('extraction-settings', 'display_node_id'))
        options_gte["add_width_thickness"]["value"] = int(config.get('extraction-settings', 'add_width_thickness'))
        options_gte["export_edge_list"]["value"] = int(config.get('extraction-settings', 'export_edge_list'))
        options_gte["export_node_positions"]["value"] = int(config.get('extraction-settings', 'export_node_positions'))
        options_gte["export_as_gexf"]["value"] = int(config.get('extraction-settings', 'export_as_gexf'))
        options_gte["export_adj_mat"]["value"] = int(config.get('extraction-settings', 'export_adj_mat'))
        options_gte["export_as_gsd"]["value"] = int(config.get('extraction-settings', 'export_as_gsd'))

        return options_gte
    except configparser.NoSectionError:
        return options_gte


def load_gtc_configs(cfg_path: str = ""):
    """Networkx Calculation Settings"""

    options_gtc = {
        "display_heatmaps": {"id": "display_heatmaps", "type": "gt-metric", "text": "Plot Heatmaps", "value": 0},
        "display_degree_histogram": {"id": "display_degree_histogram", "type": "gt-metric", "text": "Average Degree", "value": 0},
        "compute_network_diameter": {"id": "compute_network_diameter", "type": "gt-metric", "text": "Network Diameter", "value": 0},
        "compute_graph_density": {"id": "compute_graph_density", "type": "gt-metric", "text": "Graph Density", "value": 0},
        "compute_wiener_index": {"id": "compute_wiener_index", "type": "gt-metric", "text": "Wiener Index", "value": 0},
        "compute_avg_node_connectivity": {"id": "compute_avg_node_connectivity", "type": "gt-metric", "text": "Average Node Connectivity", "value": 0},
        "compute_global_efficiency": {"id": "compute_global_efficiency", "type": "gt-metric", "text": "Global Coefficient", "value": 0},
        "compute_avg_clustering_coef": {"id": "compute_avg_clustering_coef", "type": "gt-metric", "text": "Average Clustering Coefficient", "value": 0},
        "compute_assortativity_coef": {"id": "compute_assortativity_coef", "type": "gt-metric", "text": "Assortativity Coefficient", "value": 0},
        "display_betweenness_centrality_histogram": {"id": "display_betweenness_centrality_histogram", "type": "gt-metric", "text": "Betweenness Centrality", "value": 0},
        "display_closeness_centrality_histogram": {"id": "display_closeness_centrality_histogram", "type": "gt-metric", "text": "Closeness Centrality", "value": 0},
        "display_eigenvector_centrality_histogram": {"id": "display_eigenvector_centrality_histogram", "type": "gt-metric", "text": "Eigenvector Centrality", "value": 0},
        "display_edge_angle_centrality_histogram": {"id": "display_edge_angle_centrality_histogram", "type": "gt-metric", "text": "Edge Angle Centrality", "value": 0},
        "display_ohms_histogram": {"id": "display_ohms_histogram", "type": "gt-metric", "text": "Ohms Centrality", "value": 0},
        "compute_scaling_behavior": {"id": "compute_scaling_behavior", "type": "gt-metric", "text": "Scaling Behavior", "value": 0},
        "scaling_behavior_kernel_count": {"id": "scaling_behavior_kernel_count", "type": "scaling-param", "text": "No. of Kernels", "value": 10},
        "scaling_behavior_patches_per_kernel": {"id": "scaling_behavior_patches_per_kernel", "type": "scaling-param", "text": "No. of Patches per Kernel", "value": 10},
        "scaling_behavior_compute_avg": {"id": "scaling_behavior_compute_avg", "type": "scaling-param", "text": "Compute GT Averages", "value": 0},
        "scaling_behavior_power_law_fit": {"id": "scaling_behavior_power_law_fit", "type": "scaling-param", "text": "Power Law Fit", "value": 1},
        "scaling_behavior_truncated_power_law_fit": {"id": "scaling_behavior_truncated_power_law_fit", "type": "scaling-param", "text": "Truncated Power Law Fit", "value": 0},
        "scaling_behavior_log_normal_fit": {"id": "scaling_behavior_log_normal_fit", "type": "scaling-param", "text": "Log-Normal Fit", "value": 0},
        #"computing_lang": {"id": "computing_lang", "type": "gt-metric", "text": "Programming Language", "value": 'Py'}
    }

    # Load configuration from the file
    config = read_config_file(cfg_path)
    if config is None:
        return options_gtc

    try:
        options_gtc["display_heatmaps"]["value"] = int(config.get('sgt-settings', 'display_heatmaps'))
        options_gtc["display_degree_histogram"]["value"] = int(config.get('sgt-settings', 'display_degree_histogram'))
        options_gtc["display_betweenness_centrality_histogram"]["value"] = int(config.get('sgt-settings', 'display_betweenness_centrality_histogram'))
        options_gtc["display_closeness_centrality_histogram"]["value"] = int(config.get('sgt-settings', 'display_closeness_centrality_histogram'))
        options_gtc["display_eigenvector_centrality_histogram"]["value"] = int(config.get('sgt-settings', 'display_eigenvector_centrality_histogram'))
        options_gtc["compute_avg_node_connectivity"]["value"] = int(config.get('sgt-settings', 'compute_avg_node_connectivity'))
        options_gtc["compute_graph_density"]["value"] = int(config.get('sgt-settings', 'compute_graph_density'))
        options_gtc["compute_global_efficiency"]["value"] = int(config.get('sgt-settings', 'compute_global_efficiency'))
        options_gtc["compute_avg_clustering_coef"]["value"] = int(config.get('sgt-settings', 'compute_avg_clustering_coef'))
        options_gtc["compute_assortativity_coef"]["value"] = int(config.get('sgt-settings', 'compute_assortativity_coef'))
        options_gtc["compute_network_diameter"]["value"] = int(config.get('sgt-settings', 'compute_network_diameter'))
        options_gtc["compute_wiener_index"]["value"] = int(config.get('sgt-settings', 'compute_wiener_index'))
        options_gtc["display_edge_angle_centrality_histogram"]["value"] = int(config.get('sgt-settings', 'display_edge_angle_centrality_histogram'))
        options_gtc["display_ohms_histogram"]["value"] = int(config.get('sgt-settings', 'display_ohms_histogram'))
        options_gtc["compute_scaling_behavior"]["value"] = int(config.get('sgt-settings', 'compute_scaling_behavior'))
        options_gtc["scaling_behavior_compute_avg"]["value"] = int(config.get('sgt-settings', 'scaling_behavior_compute_avg'))
        options_gtc["scaling_behavior_kernel_count"]["value"] = int(config.get('sgt-settings', 'scaling_behavior_kernel_count'))
        options_gtc["scaling_behavior_patches_per_kernel"]["value"] = int(config.get('sgt-settings', 'scaling_behavior_patches_per_kernel'))
        options_gtc["scaling_behavior_power_law_fit"]["value"] = int(config.get('sgt-settings', 'scaling_behavior_power_law_fit'))
        options_gtc["scaling_behavior_truncated_power_law_fit"]["value"] = int(config.get('sgt-settings', 'scaling_behavior_truncated_power_law_fit'))
        options_gtc["scaling_behavior_log_normal_fit"]["value"] = int(config.get('sgt-settings', 'scaling_behavior_log_normal_fit'))
        # options_gtc["computing_lang"]["value"] = str(config.get('sgt-settings', 'computing_lang'))

        return options_gtc
    except configparser.NoSectionError:
        return options_gtc


def load_model_configs(cfg_path: str = ""):
    """ML/AI model settings for finding the best image filters for graph extraction"""

    options_model = {
        "find_filter_selections": {"id": "find_filter_selections", "type": "model-settings", "text": "Find Best Image Filter Combination", "value": 0},
        "find_filter_values": {"id": "find_filter_values", "type": "model-settings", "text": "Estimate Image Filter Values", "value": 0},
        "find_brightness_contrast": {"id": "find_brightness_contrast", "type": "model-settings", "text": "Estimate Brightness and Contrast Values", "value": 0},
        "max_iterations": {"id": "max_iterations", "type": "model-settings", "text": "Max. Algorithm Iterations", "value": 16},
        "genetic_alg_initial_pop": {"id": "genetic_alg_initial_pop", "type": "model-settings", "text": "Initial (Genetic Algorithm) Population Size", "value": 8},
    }

    config = read_config_file(cfg_path)
    if config is None:
        return options_model

    try:
        options_model["find_filter_selections"]["value"] = int(config.get('sgt-model', 'find_filter_selections'))
        options_model["find_filter_values"]["value"] = int(config.get('sgt-model', 'find_filter_values'))
        options_model["find_brightness_contrast"]["value"] = int(config.get('sgt-model', 'find_brightness_contrast'))
        options_model["max_iterations"]["value"] = int(config.get('sgt-model', 'max_iterations'))
        options_model["genetic_alg_initial_pop"]["value"] = int(config.get('sgt-model', 'genetic_alg_initial_pop'))

        return options_model
    except configparser.NoSectionError:
        return options_model



# SPDX-License-Identifier: GNU GPL v3
import os
import pickle
import logging
import requests
import numpy as np
from packaging import version
from typing import TYPE_CHECKING, Optional
from PySide6.QtWidgets import QApplication
from PySide6.QtCore import Property, Signal, Slot

if TYPE_CHECKING:
    # False at run time, only for a type-checker
    from _typeshed import SupportsWrite

from .tree_model import TreeModel
from .table_model import TableModel
from .checkbox_model import CheckBoxModel
from .imagegrid_model import ImageGridModel
from .qthread_worker import QThreadWorker
from ..base_worker import BaseWorker
from ..base_contoller import BaseController

from ... import __version__, __title__
from ...utils.sgt_utils import img_to_base64, verify_path
from ...imaging.image_processor import ImageProcessor, FiberNetworkBuilder, ALLOWED_IMG_EXTENSIONS, ALLOWED_GRAPH_FILE_EXTENSIONS
from ...compute.graph_analyzer import GraphAnalyzer#, COMPUTING_DEVICE


class MainController(BaseController):
    """Exposes a method to refresh the image in QML"""

    _waitFlagChanged = Signal()
    errorSignal = Signal(str)
    showAlertSignal = Signal(str, str)
    updateProgressSignal = Signal(int, str)
    taskTerminatedSignal = Signal(bool, list)
    projectOpenedSignal = Signal(str)
    changeImageSignal = Signal()
    imageChangedSignal = Signal()
    showImageHistogramSignal = Signal(bool)
    enableRectangularSelectionSignal = Signal(bool)
    showCroppingToolSignal = Signal(bool)
    showUnCroppingToolSignal = Signal(bool)
    performCroppingSignal = Signal(bool)

    def __init__(self, qml_app: QApplication):
        super().__init__()
        self._qml_app = qml_app
        self._img_loaded = False
        self._project_open = False
        self._applying_changes = False

        # Project data
        self._project_data = {"name": "", "file_path": ""}
        self._software_update = "No updates available!"

        # Initialize flags
        self._wait_flag, self._wait_flag_hist = False, False
        self._wait_msg = ""

        # Create Models
        self.imgThumbnailModel = TableModel([])
        self.imagePropsModel = TableModel([])
        self.graphPropsModel = TableModel([])
        self.graphComputeModel = TableModel([])
        self.microscopyPropsModel = CheckBoxModel([])
        self.gtcScalingModel = CheckBoxModel([])

        self.gteTreeModel = TreeModel([])
        self.gtcListModel = CheckBoxModel([])
        self.exportGraphModel = CheckBoxModel([])
        self.imgBatchModel = CheckBoxModel([])
        self.imgControlModel = CheckBoxModel([])
        self.imgBinFilterModel = CheckBoxModel([])
        self.imgFilterModel = CheckBoxModel([])
        self.imgScaleOptionModel = CheckBoxModel([])
        self.imgViewOptionModel = CheckBoxModel([])
        self.saveImgModel = CheckBoxModel([])
        self.img3dGridModel = ImageGridModel([], set([]))
        self.imgHistogramModel = ImageGridModel([], set([]))

        # Create QThreadWorker for long tasks
        self._thread_worker, self._task_worker = QThreadWorker(0, None), BaseWorker()
        self._thread_worker_hist, self._hist_worker = QThreadWorker(0, None), BaseWorker()

    def synchronize_img_models(self, sgt_obj: GraphAnalyzer):
        """
            Reload image configuration selections and controls from saved dict to QML gui_mcw after the image is loaded.

            :param sgt_obj: A GraphAnalyzer object with all saved user-selected configurations.
        """
        try:
            # Models Auto-update with saved sgt_obj configs. No need to re-assign!
            ntwk_p = sgt_obj.ntwk_p
            sel_img_batch = ntwk_p.selected_batch
            options_img = ntwk_p.image_obj.configs

            img_controls = [v for v in options_img.values() if v["type"] == "image-control"]
            bin_filters = [v for v in options_img.values() if v["type"] == "binary-filter"]
            img_filters = [v for v in options_img.values() if v["type"] == "image-filter"]
            img_properties = [v for v in options_img.values() if v["type"] == "image-property"]
            file_options = [v for v in options_img.values() if v["type"] == "file-options"]

            batch_list = [{"id": f"batch_{i}", "text": f"Image Batch {i+1}", "value": i}
                          for i in range(len(sgt_obj.ntwk_p.image_batches))]
            self.imgBatchModel.reset_data(batch_list)
            self.imgScaleOptionModel.reset_data(sel_img_batch.scaling_options)
            self.imgViewOptionModel.reset_data(sel_img_batch.view_options)

            self.imgControlModel.reset_data(img_controls)
            self.imgBinFilterModel.reset_data(bin_filters)
            self.imgFilterModel.reset_data(img_filters)
            self.microscopyPropsModel.reset_data(img_properties)
            self.saveImgModel.reset_data(file_options)
        except Exception as err:
            logging.exception("Fatal Error: %s", err, extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Fatal Error", "Error re-loading image configurations! Close app and try again.")

    def synchronize_graph_models(self, sgt_obj: GraphAnalyzer):
        """
            Reload graph configuration selections and controls from saved dict to QML gui_mcw.
        Args:
            sgt_obj: a GraphAnalyzer object with all saved user-selected configurations.

        Returns:

        """
        try:
            # Models Auto-update with saved sgt_obj configs. No need to re-assign!
            ntwk_p = sgt_obj.ntwk_p
            sel_img_batch = ntwk_p.selected_batch
            graph_obj = ntwk_p.graph_obj
            option_gte = graph_obj.configs
            options_gtc = sgt_obj.configs

            graph_options = [v for v in option_gte.values() if v["type"] == "graph-extraction"]
            file_options = [v for v in option_gte.values() if v["type"] == "file-options"]
            compute_options = [v for v in options_gtc.values() if v["type"] == "gt-metric"]
            scaling_options = [v for v in options_gtc.values() if v["type"] == "scaling-param"]

            self.gteTreeModel.reset_data(graph_options)
            self.exportGraphModel.reset_data(file_options)
            self.gtcListModel.reset_data(compute_options)
            self.gtcScalingModel.reset_data(scaling_options)

            self.imagePropsModel.reset_data(sel_img_batch.props)
            self.graphPropsModel.reset_data(graph_obj.props)
            self.graphComputeModel.reset_data(sgt_obj.props)
        except Exception as err:
            logging.exception("Fatal Error: %s", err, extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Fatal Error", "Error re-loading image configurations! Close app and try again.")

    def delete_sgt_object(self, index=None):
        """
        Delete SGT Obj stored at the specified index (if not specified, get the current index).
        """
        del_index = index if index is not None else self._selected_sgt_obj_index
        if 0 <= del_index < len(self._sgt_objs):  # Check if the index exists
            keys_list = list(self._sgt_objs.keys())
            key_at_del_index = keys_list[self._selected_sgt_obj_index]
            # Delete the object at index
            del self._sgt_objs[key_at_del_index]
            # Update Data
            img_list, img_cache = self.get_thumbnail_list()
            self.imgThumbnailModel.update_data(img_list, img_cache)
            self.imagePropsModel.reset_data([])
            self.graphPropsModel.reset_data([])
            self.graphComputeModel.reset_data([])
            self._selected_sgt_obj_index = 0
            self.load_image(reload_thumbnails=True)
            self.imageChangedSignal.emit()

    def save_project_data(self):
        """
        A handler function that handles saving project data.
        Returns: True if successful, False otherwise.

        """
        if not self._project_open:
            return False
        try:
            file_path = self._project_data["file_path"]
            with open(file_path, 'wb') as project_file:  # type: Optional[SupportsWrite[bytes]]
                pickle.dump(self._sgt_objs, project_file)
            return True
        except Exception as err:
            logging.exception("Project Saving Error: %s", err, extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Save Error", "Unable to save project data. Close app and try again.")
            return False

    def get_thumbnail_list(self):
        """
        Get names and base64 data of images to be used in Project List thumbnails.
        """
        keys_list = list(self._sgt_objs.keys())
        if len(keys_list) <= 0:
            return None, None
        item_data = []
        image_cache = {}
        for key in keys_list:
            item_data.append([key])  # Store the key
            sgt_obj = self._sgt_objs[key]
            if sgt_obj.ntwk_p.selected_batch.is_graph_only:
                empty_cv = np.ones((256, 256), dtype=np.uint8) * 255
                img_cv = empty_cv if sgt_obj.ntwk_p.graph_obj.img_ntwk is None else sgt_obj.ntwk_p.graph_obj.img_ntwk
            else:
                img_cv = sgt_obj.ntwk_p.image_2d
            base64_data = img_to_base64(img_cv)
            image_cache[key] = base64_data  # Store base64 string
        return item_data, image_cache

    def get_selected_images(self):
        """
        Get selected images from a specific image batch.
        """
        sgt_obj = self.get_selected_sgt_obj()
        ntwk_p = sgt_obj.ntwk_p
        return ntwk_p.selected_images

    def _handle_progress_update(self, progress_val: int, msg: str) -> None:
        """
        Handler function for progress updates for ongoing tasks.
        Args:
            progress_val: Progress value, range is 0-100%.
            msg: Progress message to be displayed.

        Returns:

        """

        if 0 <= progress_val <= 100:
            self.updateProgressSignal.emit(progress_val, msg)
            logging.info(f"{progress_val}%: {msg}", extra={'user': 'SGT Logs'})
        elif progress_val > 100:
            self.updateProgressSignal.emit(progress_val, msg)
            logging.info(f"{msg}", extra={'user': 'SGT Logs'})
        else:
            logging.exception("Error: %s", msg, extra={'user': 'SGT Logs'})
            self.errorSignal.emit(msg)

    def _handle_finished(self, success_val: bool, result: None | list | ImageProcessor | FiberNetworkBuilder | GraphAnalyzer) -> None:
        """
        Handler function for sending updates/signals on termination of tasks.
        Args:
            success_val:
            result:

        Returns:

        """
        self._stop_wait()
        if not success_val:
            if type(result) is list:
                logging.info(result[0] + ": " + result[1], extra={'user': 'SGT Logs'})
                self.taskTerminatedSignal.emit(success_val, result)
            # elif type(result) is GraphAnalyzer:
            #    pdf_saved = GraphAnalyzer.write_to_pdf(result, self._handle_progress_update)
            #    if pdf_saved:
            #        self._handle_finished(True, result)
        else:
            if type(result) is str:
                # Saving files to Output Folder
                self._handle_progress_update(100, "Files Saved!")
                self.taskTerminatedSignal.emit(success_val, ["Files Saved", result])
            elif type(result) is ImageProcessor:
                self._handle_progress_update(100, "Graph extracted successfully!")
                # Update Compute properties
                self.changeImageSignal.emit()
                self.synchronize_graph_models(self.get_selected_sgt_obj())
                # Send task termination signal to QML
                self.taskTerminatedSignal.emit(success_val, [])
            elif type(result) is GraphAnalyzer:
                self._handle_progress_update(100, "GT PDF successfully generated!")
                # Update Compute properties
                self.synchronize_graph_models(self.get_selected_sgt_obj())
                # Send task termination signal to QML
                self.taskTerminatedSignal.emit(True, ["GT calculations completed", "The image's GT parameters have been "
                                                                                "calculated. Check out generated PDF in "
                                                                                "'Output Dir'."])
            elif type(result) is dict:
                self._handle_progress_update(100, "All GT PDF successfully generated!")
                # Update Compute properties
                self.synchronize_graph_models(self.get_selected_sgt_obj())
                # Send task termination signal to QML
                self.taskTerminatedSignal.emit(True, ["All GT calculations completed", "GT parameters of all "
                                                                                    "images have been calculated. Check "
                                                                                    "out all the generated PDFs in "
                                                                                    "'Output Dir'."])
            elif type(result) is list:
                # Image histogram calculated
                self._wait_flag_hist = False
                if len(self._sgt_objs) > 0:
                    sgt_obj = self.get_selected_sgt_obj()
                    sel_img_batch = sgt_obj.ntwk_p.selected_batch
                    self.imgHistogramModel.reset_data(result, sel_img_batch.selected_images_idx)
            else:
                self.taskTerminatedSignal.emit(success_val, [])

            # Auto-save changes to the project data file
            if len(self._sgt_objs.items()) <= 10:
                self.save_project_data()

    def _start_wait(self, msg: str = "please wait..."):
        """Activate the wait flag and send a wait signal."""
        self._wait_flag = True
        self._wait_msg = msg
        self._waitFlagChanged.emit()

    def _stop_wait(self):
        """Deactivate the wait flag and send a wait signal."""
        self._wait_flag = False
        self._wait_msg = ""
        self._waitFlagChanged.emit()

    # --- Properties exposed to QML because of "notify" ---
    @Property(bool, notify=_waitFlagChanged)
    def wait(self):
        return self._wait_flag

    @Property(str, notify=_waitFlagChanged)
    def wait_text(self):
        return self._wait_msg

    @Slot(result=str)
    def get_sgt_title(self):
        return f"{__title__}"

    @Slot(result=str)
    def get_sgt_version(self):
        """"""
        # return f"{__title__} v{__version__}, Computing: {COMPUTING_DEVICE}"
        return f"v{__version__}"

    @Slot(result=str)
    def get_software_download_details(self):
        return self._software_update

    @Slot(result=str)
    def get_about_details(self):
        about_app = (
            "<html>"
            "<p>"
            "A software tool for performing Graph Theory analysis on <br>microscopy images. This is a modified version "
            "of StructuralGT <br>initially proposed by Drew A. Vecchio,<br>"
            "<b>DOI:</b> <a href='https://pubs.acs.org/doi/10.1021/acsnano.1c04711'>10.1021/acsnano.1c04711</a>"
            "</p><p>"
            "<b>Main Contributors:</b><br>"
            "<table border='0.5' cellspacing='0' cellpadding='4'>"
            # "<tr><th>Name</th><th>Email</th></tr>"
            "<tr><td>Dickson Owuor</td><td>owuor@umich.edu</td></tr>"
            "<tr><td>Nicolas Kotov</td><td>kotov@umich.edu</td></tr>"
            "<tr><td>Alain Kadar</td><td>alaink@umich.edu</td></tr>"
            "<tr><td>Xiong Ye Xiao</td><td>xiongyex@usc.edu</td></tr>"
            "<tr><td>Kotov Lab</td><td></td></tr>"
            "<tr><td>COMPASS</td><td></td></tr>"
            "</table></p><p><br><br>"
            "<b>Documentation:</b> <a href='https://structural-gt.readthedocs.io'>structural-gt.readthedocs.io</a>"
            "<br>"
            f"<b> Version: </b> {self.get_sgt_version()}<br>"
            "<b>License:</b> GPL GNU v3"
            "</p><p>"
            "Copyright (C) 2018-2025<br>The Regents of the University of Michigan."
            "</p>"
            "</html>")
        return about_app

    @Slot(result=bool)
    def check_for_updates(self):
        """"""
        github_url = "https://raw.githubusercontent.com/owuordickson/structural-gt/refs/heads/main/src/sgtlib/__init__.py"

        try:
            response = requests.get(github_url, timeout=5)
            response.raise_for_status()
        except requests.RequestException as e:
            self._software_update = f"Error checking for updates: {e}"
            return False

        remote_version = None
        for line in response.text.splitlines():
            if line.strip().startswith("__install_version__"):
                try:
                    remote_version = line.split("=")[1].strip().strip("\"'")
                    break
                except IndexError:
                    self._software_update = "Could not connect to server!"
                    return False

        if not remote_version:
            self._software_update = "Could not find the new version!"
            return False

        new_version = version.parse(remote_version)
        current_version = version.parse(__version__)
        if new_version > current_version:
            # https://github.com/owuordickson/structural-gt/releases/tag/v3.3.5
            self._software_update = (
                "New version available!<br>"
                f"Download via this <a href='https://github.com/owuordickson/structural-gt/releases/tag/v{remote_version}'>link</a>"
            )
            return True
        else:
            self._software_update = "No updates available."
            return False

    @Slot(str, result=str)
    def get_file_extensions(self, option):
        if option == "img":
            pattern_string = ' '.join(ALLOWED_IMG_EXTENSIONS)
            return f"Image files ({pattern_string})"
        if option == "graph":
            pattern_string = ' '.join(ALLOWED_GRAPH_FILE_EXTENSIONS)
            return f"Graph files ({pattern_string})"
        elif option == "proj":
            return "Project files (*.sgtproj)"
        else:
            return ""

    @Slot(result=str)
    def get_pixmap(self):
        """Returns the URL that QML should use to load the image"""
        curr_img_view = np.random.randint(0, 4)
        unique_num = self._selected_sgt_obj_index + curr_img_view + np.random.randint(low=21, high=1000)
        return "image://imageProvider/" + str(unique_num)

    @Slot(result=bool)
    def is_img_3d(self):
        sgt_obj = self.get_selected_sgt_obj()
        if sgt_obj is None:
            return False
        sel_img_batch = sgt_obj.ntwk_p.selected_batch
        is_3d = not sel_img_batch.is_2d
        return is_3d

    @Slot(result=int)
    def get_selected_img_batch(self):
        try:
            sgt_obj = self.get_selected_sgt_obj()
            return sgt_obj.ntwk_p.selected_batch_index
        except AttributeError:
            logging.exception("No image added! Please add at least one image.", extra={'user': 'SGT Logs'})
            return 0

    @Slot(result=str)
    def get_img_nav_location(self):
        return f"{(self._selected_sgt_obj_index + 1)} / {len(self._sgt_objs)}"

    @Slot(result=str)
    def get_output_dir(self):
        sgt_obj = self.get_selected_sgt_obj()
        if sgt_obj is None:
            return ""
        return f"{sgt_obj.ntwk_p.output_dir}"

    @Slot(result=bool)
    def get_auto_scale(self):
        return self._allow_auto_scale

    @Slot(int)
    def delete_selected_thumbnail(self, img_index):
        """Delete the selected image from the list."""
        self.delete_sgt_object(img_index)

    @Slot(str)
    def set_output_dir(self, folder_path):
        self.update_output_dir(folder_path)
        self.imageChangedSignal.emit()

    @Slot(bool)
    def set_auto_scale(self, auto_scale):
        """Set the auto-scale parameter for each image."""
        self._allow_auto_scale = auto_scale

    @Slot(int)
    def select_img_batch(self, batch_index=-1):
        if batch_index < 0:
            return

        try:
            sgt_obj = self.get_selected_sgt_obj()
            sgt_obj.ntwk_p.select_image_batch(batch_index)
            # Load the SGT Object data of the selected image
            self.synchronize_img_models(sgt_obj)
            self.synchronize_graph_models(self.get_selected_sgt_obj())
            # Load the selected image into the view
            self.changeImageSignal.emit()
        except Exception as err:
            logging.exception("Batch Change Error: %s", err, extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Image Batch Error", f"Error encountered while trying to access batch "
                                                           f"{batch_index}. Restart app and try again.")

    @Slot(int, bool)
    def toggle_selected_batch_image(self, img_index, selected):
        sgt_obj = self.get_selected_sgt_obj()
        sel_img_batch = sgt_obj.ntwk_p.selected_batch
        if selected:
            sel_img_batch.selected_images_idx.add(img_index)
        else:
            sel_img_batch.selected_images_idx.discard(img_index)
        self.changeImageSignal.emit()

    @Slot(bool)
    def reload_graph_image(self, only_giant_graph=False):
        sgt_obj = self.get_selected_sgt_obj()
        sel_img_batch = sgt_obj.ntwk_p.selected_batch
        sgt_obj.ntwk_p.draw_graph_image(sel_img_batch, show_giant_only=only_giant_graph)
        self.changeImageSignal.emit()

    @Slot()
    def load_graph_simulation(self):
        """Render and visualize OVITO graph network simulation."""
        try:
            # Import libraries
            from ovito import scene
            from ovito.vis import Viewport
            from ovito.io import import_file
            from ovito.gui import create_qwidget

            # Clear any existing scene
            for p_line in list(scene.pipelines):
                p_line.remove_from_scene()

            # Create OVITO data pipeline
            sgt_obj = self.get_selected_sgt_obj()
            h, w = sgt_obj.ntwk_p.graph_obj.img_ntwk.shape[:2]
            pipeline = import_file(sgt_obj.ntwk_p.graph_obj.gsd_file)
            pipeline.add_to_scene()

            vp = Viewport(type=Viewport.Type.Perspective, camera_dir=(2, 1, -1))
            vp.zoom_all((w, h))  # width, height

            ovito_widget = create_qwidget(vp, parent=self._qml_app.activeWindow())
            ovito_widget.setFixedSize(w, h)
            ovito_widget.show()
        except Exception as e:
            print("Graph Simulation Error:", e)

    @Slot(int)
    def load_image(self, index=None, reload_thumbnails=False):
        try:
            if index is not None:
                if index == self._selected_sgt_obj_index:
                    return
                else:
                    self._selected_sgt_obj_index = index

            if reload_thumbnails:
                # Update the thumbnail list data (delete/add image)
                img_list, img_cache = self.get_thumbnail_list()
                self.imgThumbnailModel.update_data(img_list, img_cache)

            # Load the SGT Object data of the selected image
            self.synchronize_img_models(self.get_selected_sgt_obj())
            self.synchronize_graph_models(self.get_selected_sgt_obj())
            self.imgThumbnailModel.set_selected(self._selected_sgt_obj_index)
            # Load the selected image into the view
            self.changeImageSignal.emit()
        except Exception as err:
            self.delete_sgt_object()
            self._selected_sgt_obj_index = 0
            logging.exception("Image Loading Error: %s", err, extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Image Error", "Error loading image. Try again.")

    @Slot(result=bool)
    def load_prev_image(self):
        """Load the previous image in the list into view."""
        if self._selected_sgt_obj_index > 0:
            prev_img_idx = self._selected_sgt_obj_index - 1
            self.load_image(index=prev_img_idx)
            return True
        return False

    @Slot(result=bool)
    def load_next_image(self):
        """Load the next image in the list into view."""
        if self._selected_sgt_obj_index < (len(self._sgt_objs) - 1):
            next_img_idx = self._selected_sgt_obj_index + 1
            self.load_image(index=next_img_idx)
            return True
        return False

    @Slot(str)
    def apply_changes(self, view: str=""):
        """Retrieve changes made by the user and apply to image/graph."""
        if not self._applying_changes:  # Disallow concurrent changes
            # print(f"change to {view}")
            self._applying_changes = True
            if view != "":
                sgt_obj = self.get_selected_sgt_obj()
                sgt_obj.ntwk_p.selected_batch_view = view
            self.changeImageSignal.emit()

    @Slot()
    def compute_img_histogram(self):
        """Calculate the histogram of the image."""
        if self._wait_flag_hist:
            return

        self.showImageHistogramSignal.emit(True)
        self._hist_worker = BaseWorker()
        try:
            self._wait_flag_hist = True
            sgt_obj = self.get_selected_sgt_obj()
            self._thread_worker_hist = QThreadWorker(func=self._hist_worker.task_calculate_img_histogram, args=(sgt_obj.ntwk_p,))
            self._hist_worker.taskFinishedSignal.connect(self._handle_finished)
            self._thread_worker_hist.start()
        except Exception as err:
            self._wait_flag_hist = False
            logging.exception("Histogram Calculation Error: %s", err, extra={'user': 'SGT Logs'})
            self._handle_finished(False, ["Histogram Calculation Failed", "Unable to calculate image histogram!"])

    @Slot()
    def apply_img_scaling(self):
        """Retrieve settings from the model and send to Python."""
        try:
            self.set_auto_scale(True)
            sgt_obj = self.get_selected_sgt_obj()
            sgt_obj.ntwk_p.auto_scale = self._allow_auto_scale
            sgt_obj.ntwk_p.apply_img_scaling()
            self.changeImageSignal.emit()
        except Exception as err:
            logging.exception("Apply Image Scaling: " + str(err), extra={'user': 'SGT Logs'})
            self.taskTerminatedSignal.emit(False, ["Unable to Rescale Image", "Error while tying to re-scale "
                                                                              "image. Try again."])

    @Slot()
    def export_graph_to_file(self):
        """Export graph data and save as a file."""
        if self._wait_flag:
            logging.info("Please Wait: Another Task Running!", extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Please Wait", "Another Task Running!")
            return

        self._handle_progress_update(0, "Exporting Graph Data...")
        self._task_worker = BaseWorker()
        try:
            if self.get_selected_sgt_obj().ntwk_p.selected_batch.is_graph_only:
                return

            self._handle_progress_update(20, "Exporting Graph Data...")
            self._start_wait()
            sgt_obj = self.get_selected_sgt_obj()

            self._thread_worker = QThreadWorker(func=self._task_worker.task_export_graph, args=(sgt_obj.ntwk_p,))
            self._task_worker.taskFinishedSignal.connect(self._handle_finished)
            self._thread_worker.start()
        except Exception as err:
            logging.exception("Unable to Export Graph: " + str(err), extra={'user': 'SGT Logs'})
            self.taskTerminatedSignal.emit(False, ["Unable to Export Graph", "Error exporting graph to file. Try again."])

    @Slot()
    def save_img_files(self):
        """Retrieve and save images to the file."""
        if self._wait_flag:
            logging.info("Please Wait: Another Task Running!", extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Please Wait", "Another Task Running!")
            return

        self._handle_progress_update(0, "Saving Images...")
        self._task_worker = BaseWorker()
        try:
            if self.get_selected_sgt_obj().ntwk_p.selected_batch.is_graph_only:
                return

            self._handle_progress_update(10, "Saving Images...")
            sel_images = self.get_selected_images()
            for val in self.saveImgModel.list_data:
                for img in sel_images:
                    img.configs[val["id"]]["value"] = val["value"]

            self._handle_progress_update(20, "Saving Images...")
            self._start_wait()
            sgt_obj = self.get_selected_sgt_obj()

            self._thread_worker = QThreadWorker(func=self._task_worker.task_save_images, args=(sgt_obj.ntwk_p,))
            self._task_worker.taskFinishedSignal.connect(self._handle_finished)
            self._thread_worker.start()
        except Exception as err:
            logging.exception("Unable to Save Image Files: " + str(err), extra={'user': 'SGT Logs'})
            self.taskTerminatedSignal.emit(False,
                                           ["Unable to Save Image Files", "Error saving images to file. Try again."])

    @Slot()
    def run_extract_graph(self):
        """Retrieve settings from the model and send to Python."""

        if self._wait_flag:
            logging.info("Please Wait: Another Task Running!", extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Please Wait", "Another Task Running!")
            return

        self._task_worker = BaseWorker()
        try:
            self._start_wait()
            sgt_obj = self.get_selected_sgt_obj()

            self._thread_worker = QThreadWorker(func=self._task_worker.task_extract_graph, args=(sgt_obj.ntwk_p,))
            self._task_worker.inProgressSignal.connect(self._handle_progress_update)
            self._task_worker.taskFinishedSignal.connect(self._handle_finished)
            self._thread_worker.start()
        except Exception as err:
            self._stop_wait()
            logging.exception("Graph Extraction Error: %s", err, extra={'user': 'SGT Logs'})
            self._handle_progress_update(-1, "Fatal error occurred! Close the app and try again.")
            self._handle_finished(False, ["Graph Extraction Error",
                                                          "Fatal error while trying to extract graph. "
                                                          "Close the app and try again."])

    @Slot()
    def run_graph_analyzer(self):
        """Retrieve settings from the model and send to Python."""
        if self._wait_flag:
            logging.info("Please Wait: Another Task Running!", extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Please Wait", "Another Task Running!")
            return

        self._task_worker = BaseWorker()
        try:
            self._start_wait()
            sgt_obj = self.get_selected_sgt_obj()

            self._thread_worker = QThreadWorker(func=self._task_worker.task_compute_gt, args=(sgt_obj,))
            self._task_worker.inProgressSignal.connect(self._handle_progress_update)
            self._task_worker.taskFinishedSignal.connect(self._handle_finished)
            self._thread_worker.start()
        except Exception as err:
            self._stop_wait()
            logging.exception("GT Computation Error: %s", err, extra={'user': 'SGT Logs'})
            self._handle_progress_update(-1, "Fatal error occurred! Close the app and try again.")
            self._handle_finished(False, ["GT Computation Error",
                                                          "Fatal error while trying calculate GT parameters. "
                                                          "Close the app and try again."])

    @Slot()
    def run_multi_graph_analyzer(self):
        """"""
        if self._wait_flag:
            logging.info("Please Wait: Another Task Running!", extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Please Wait", "Another Task Running!")
            return

        self._task_worker = BaseWorker()
        try:
            self._start_wait()

            # Update Configs
            self.replicate_sgt_configs()

            self._thread_worker = QThreadWorker(func=self._task_worker.task_compute_multi_gt, args=(self._sgt_objs,))
            self._task_worker.inProgressSignal.connect(self._handle_progress_update)
            self._task_worker.taskFinishedSignal.connect(self._handle_finished)
            self._thread_worker.start()
        except Exception as err:
            self._stop_wait()
            logging.exception("GT Computation Error: %s", err, extra={'user': 'SGT Logs'})
            self._handle_progress_update(-1, "Fatal error occurred! Close the app and try again.")
            self._handle_finished(False, ["GT Computation Error",
                                                          "Fatal error while trying calculate GT parameters. "
                                                          "Close the app and try again."])

    @Slot(result=bool)
    def run_save_project(self):
        """"""
        if self._wait_flag:
            logging.info("Please Wait: Another Task Running!", extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Please Wait", "Another Task Running!")
            return False

        self._start_wait()
        success_val = self.save_project_data()
        self._stop_wait()
        return success_val

    @Slot(result=bool)
    def enable_img_controls(self):
        """Enable image controls."""
        if len(self._sgt_objs) <= 0:
            return False

        sgt_obj = self.get_selected_sgt_obj()
        if sgt_obj is None:
            return False
        return not sgt_obj.ntwk_p.selected_batch.is_graph_only

    @Slot(result=bool)
    def display_image(self):
        return self._img_loaded

    @Slot(result=bool)
    def display_graph(self):
        if len(self._sgt_objs) <= 0:
            return False

        sgt_obj = self.get_selected_sgt_obj()
        if sgt_obj is None:
            return False

        if sgt_obj.ntwk_p.graph_obj.img_ntwk is None:
            return False

        if sgt_obj.ntwk_p.selected_batch_view == "graph":
            return True
        return False

    @Slot(result=bool)
    def image_batches_exist(self):
        if not self._img_loaded:
            return False

        sgt_obj = self.get_selected_sgt_obj()
        batch_count = len(sgt_obj.ntwk_p.image_batches)
        batches_exist = True if batch_count > 1 else False
        return batches_exist

    @Slot(result=bool)
    def is_project_open(self):
        return self._project_open

    @Slot(result=bool)
    def is_task_running(self):
        return self._wait_flag

    @Slot(bool)
    def show_cropping_tool(self, allow_cropping):
        self.showCroppingToolSignal.emit(allow_cropping)

    @Slot(bool)
    def perform_cropping(self, allowed):
        self.performCroppingSignal.emit(allowed)

    @Slot( int, int, int, int, int, int)
    def crop_image(self, x, y, crop_width, crop_height, qimg_width, qimg_height):
        """Crop image using PIL and save it."""
        try:
            sgt_obj = self.get_selected_sgt_obj()
            sgt_obj.ntwk_p.crop_image(x, y, crop_width, crop_height, qimg_width, qimg_height)

            # Emit signal to update UI with new image
            self.changeImageSignal.emit()
            self.showCroppingToolSignal.emit(False)
            self.showUnCroppingToolSignal.emit(True)
        except Exception as err:
            logging.exception("Cropping Error: %s", err, extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Cropping Error", "Error occurred while cropping image. Close the app and try again.")

    @Slot(bool)
    def undo_cropping(self, undo: bool = True):
        if undo:
            sgt_obj = self.get_selected_sgt_obj()
            sgt_obj.ntwk_p.undo_cropping()

            # Emit signal to update UI with new image
            self.changeImageSignal.emit()
            self.showUnCroppingToolSignal.emit(False)

    @Slot(bool)
    def enable_rectangular_selection(self, enabled):
        self.enableRectangularSelectionSignal.emit(enabled)

    @Slot(result=bool)
    def enable_prev_nav_btn(self):
        if (self._selected_sgt_obj_index == 0) or self.is_task_running():
            return False
        else:
            return True

    @Slot(result=bool)
    def enable_next_nav_btn(self):
        if (self._selected_sgt_obj_index == (len(self._sgt_objs) - 1)) or self.is_task_running():
            return False
        else:
            return True

    @Slot(str, result=bool)
    def upload_graph_file(self, file_path):
        """Verify and validate the file path, use it to create a new SGT Object and load it into the view."""
        is_successful = self.add_graph(file_path)
        if is_successful:
            self.synchronize_img_models(self.get_selected_sgt_obj())
            self.synchronize_graph_models(self.get_selected_sgt_obj())
            self.load_image(reload_thumbnails=True)
        return is_successful

    @Slot(str, result=bool)
    def upload_single_image(self, img_path):
        """Verify and validate the image path, use it to create an SGT object and load it in view."""
        is_successful = self.add_single_image(img_path)
        if is_successful:
            self.synchronize_img_models(self.get_selected_sgt_obj())
            self.synchronize_graph_models(self.get_selected_sgt_obj())
            self.load_image(reload_thumbnails=True)
        return is_successful

    @Slot(str, result=bool)
    def upload_multiple_images(self, img_dir_path):
        """
        Verify and validate multiple image paths, use each to create an SGT object, then load the last one in view.
        """
        is_successful = self.add_multiple_images(img_dir_path)
        if is_successful:
            self.synchronize_img_models(self.get_selected_sgt_obj())
            self.synchronize_graph_models(self.get_selected_sgt_obj())
            self.load_image(reload_thumbnails=True)
        return is_successful

    @Slot(str, str, result=bool)
    def create_sgt_project(self, proj_name, dir_path):
        """Creates a '.sgtproj' inside the selected directory"""

        self._project_open = False
        success, result = verify_path(dir_path)
        if success:
            dir_path = result
        else:
            logging.info(result, extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("File/Directory Error", result)
            return False

        proj_name += '.sgtproj'
        proj_path = os.path.join(str(dir_path), proj_name)

        try:
            if os.path.exists(proj_path):
                logging.info(f"Project '{proj_name}' already exists.", extra={'user': 'SGT Logs'})
                self.showAlertSignal.emit("Project Error", f"Error: Project '{proj_name}' already exists.")
                return False

            # Open the file in the 'write' mode ('w').
            # This will create the file if it doesn't exist
            with open(proj_path, 'w'):
                pass  # Do nothing, just create the file (updates will be done automatically/dynamically)

            # Update and notify QML
            self._project_data["name"] = proj_name
            self._project_data["file_path"] = proj_path
            self._project_open = True
            self.projectOpenedSignal.emit(proj_name)
            logging.info(f"File '{proj_name}' created successfully in '{dir_path}'.", extra={'user': 'SGT Logs'})
            return True
        except Exception as err:
            logging.exception("Create Project Error: %s", err, extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Create Project Error", "Failed to create SGT project. Close the app and try again.")
            return False

    @Slot(str, result=bool)
    def open_sgt_project(self, sgt_path):
        """Opens and loads the SGT project from the '.sgtproj' file"""
        if self._wait_flag:
            logging.info("Please Wait: Another Task Running!", extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Please Wait", "Another Task Running!")
            return False

        try:
            self._start_wait()
            self._project_open = False
            # Verify the path
            success, result = verify_path(sgt_path)
            if success:
                sgt_path = result
            else:
                logging.info(result, extra={'user': 'SGT Logs'})
                self.showAlertSignal.emit("File/Directory Error", result)
                self._stop_wait()
                return False
            img_dir, proj_name = os.path.split(str(sgt_path))

            # Read and load project data and SGT objects
            with open(str(sgt_path), 'rb') as sgt_file:
                self._sgt_objs = pickle.load(sgt_file)

            if self._sgt_objs:
                key_list = list(self._sgt_objs.keys())
                for key in key_list:
                    self._sgt_objs[key].ntwk_p.output_dir = img_dir

            # Update and notify QML
            self._project_data["name"] = proj_name
            self._project_data["file_path"] = str(sgt_path)
            self._stop_wait()
            self._project_open = True
            self.projectOpenedSignal.emit(proj_name)

            # Load Image to GUI - activates QML
            self.load_image(reload_thumbnails=True)
            logging.info(f"File '{proj_name}' opened successfully in '{sgt_path}'.", extra={'user': 'SGT Logs'})
            return True
        except Exception as err:
            self._stop_wait()
            logging.exception("Project Opening Error: %s", err, extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Open Project Error", "Unable to open .sgtproj file! Try again. If the "
                                                            "issue persists, the file may be corrupted or incompatible. "
                                                            "Consider restoring from a backup or contacting support for "
                                                            "assistance.")
            return False



# SPDX-License-Identifier: GNU GPL v3

"""
Processes 2D or 3D images and generate a fiber graph network.
"""

import re
import os
import cv2
# import pydicom
import logging
import numpy as np
# import nibabel as nib
from PIL import Image
from cv2.typing import MatLike
from dataclasses import dataclass
from collections import defaultdict

from ..imaging.base_image import BaseImage
from ..models.filter_env import FilterSearchSpace
from ..networks.fiber_network import FiberNetworkBuilder
from ..utils.progress_update import ProgressUpdate
from ..utils.config_loader import load_model_configs
from ..utils.sgt_utils import plot_to_opencv, AbortException
from ..models.filter_env import sgt_genetic_algorithm, sgt_hill_climbing_algorithm

logger = logging.getLogger("SGT App")

Image.MAX_IMAGE_PIXELS = None  # Disable limit on maximum image size
ALLOWED_IMG_EXTENSIONS = ('*.jpg', '*.png', '*.jpeg', '*.bmp', '*.tif', '*.tiff', '*.qptiff')
ALLOWED_3D_IMG_EXTENSIONS = ('*.qptiff', '*.nii', '*.nii.gz', '*.dcm')
ALLOWED_GRAPH_FILE_EXTENSIONS = ('*.csv', '*.gsd')


class ImageProcessor(ProgressUpdate):
    """
    A class for processing and preparing 2D or 3D microscopy images for building a fiber graph network.

    Args:
        img_path (str): input image path
        out_dir (str): directory path for storing results.
    """

    @dataclass
    class ImageBatch:
        """A class for storing image batch data."""
        numpy_image: np.ndarray
        images: list[BaseImage]
        graph_obj: FiberNetworkBuilder
        shape: tuple
        props: list
        is_2d: bool
        is_graph_only: bool
        scale_factor: float
        scaling_options: list[dict]
        selected_images_idx: set
        view_options: list[dict]

    def __init__(self, img_path, out_dir, cfg_file="", graph_file="", auto_scale=True):
        """
        A class for processing and preparing microscopy images for building a fiber graph network.

        Args:
            img_path (str | list): input image path
            out_dir (str): directory path for storing results
            cfg_file (str): configuration file path
            graph_file (str): graph file path (when creating the graph from CSV/GSD data)
            auto_scale (bool): whether to automatically scale the image

        >>>
        >>> i_path = "path/to/image"
        >>> cfg_path = "path/to/sgt_configs.ini"
        >>>
        >>> ntwk_p, img_file = ImageProcessor.from_image_file(i_path, config_file=cfg_path)
        >>> ntwk_p.apply_img_filters()
        """
        super(ImageProcessor, self).__init__()
        self._configs = load_model_configs(cfg_file)
        self._img_path: str = img_path if type(img_path) is str else img_path[0]
        self._output_dir: str = out_dir
        self._config_file: str = cfg_file
        self._graph_file: str = graph_file
        self._auto_scale: bool = auto_scale
        self._image_batches: list[ImageProcessor.ImageBatch] = []
        self._selected_batch_index: int = 0
        # self._initialize_image_batches(self._load_img_from_file(img_path))

    @property
    def configs(self):
        """Returns the configuration (ML model) settings for the image processor."""
        return self._configs

    @configs.setter
    def configs(self, configs):
        """Sets the configuration (ML model) settings for the image processor."""
        self._configs = configs

    @property
    def img_path(self) -> str:
        """Returns the input image path."""
        return self._img_path

    @property
    def output_dir(self) -> str:
        """Returns the output directory path for storing results."""
        return self._output_dir

    @output_dir.setter
    def output_dir(self, folder: str):
        """Sets the output directory path for storing results."""
        self._output_dir = folder

    @property
    def config_file(self) -> str:
        """Returns the configuration file path (usually sgt_configs.ini)."""
        return self._config_file

    @property
    def graph_file(self) -> str:
        """Returns the graph file path."""
        return self._graph_file

    @property
    def auto_scale(self) -> bool:
        """Returns whether to automatically scale the image."""
        return self._auto_scale

    @auto_scale.setter
    def auto_scale(self, value: bool):
        self._auto_scale = value

    @property
    def image_batches(self) -> list["ImageProcessor.ImageBatch"]:
        """Returns a list of ImageBatch objects."""
        return self._image_batches

    @property
    def selected_batch_index(self) -> int:
        """Returns the selected batch index."""
        return self._selected_batch_index

    @property
    def selected_batch(self):
        """
        Retrieved data of the current selected batch.
        """
        return self._image_batches[self._selected_batch_index]

    @property
    def selected_batch_view(self):
        """Gets the current image batch view."""
        sel_img_batch = self.selected_batch
        for view_dict in sel_img_batch.view_options:
            if view_dict["value"] == 1:
                return view_dict["dataValue"]
        return sel_img_batch.view_options[0]["dataValue"]

    @selected_batch_view.setter
    def selected_batch_view(self, value):
        """Sets the current image batch view."""
        sel_img_batch = self.selected_batch
        for view_dict in sel_img_batch.view_options:
            view_dict["value"] = 1 if value == view_dict["dataValue"] else 0

    @property
    def selected_images(self) -> list[BaseImage]:
        """Returns a list of selected images."""
        sel_img_batch = self.selected_batch
        if sel_img_batch.is_graph_only:
            return []
        sel_images = [sel_img_batch.images[i] for i in sel_img_batch.selected_images_idx]
        return sel_images

    @property
    def image_obj(self) -> BaseImage:
        """Returns the first image (2D) object/instance in the batch."""
        sel_img_batch = self.selected_batch
        first_index = next(iter(sel_img_batch.selected_images_idx), None)  # 1st selected image
        first_index = first_index if first_index is not None else 0  # first image if None
        return sel_img_batch.images[first_index]

    @property
    def image_obj_3d(self) -> list[BaseImage]:
        """Returns the full image list (3D) BaseImage objects/instances in the batch."""
        return self.selected_batch.images

    @property
    def graph_obj(self):
        """Returns the NetworkX graph extracted from the image."""
        return self.selected_batch.graph_obj

    @property
    def image_histogram(self) -> MatLike:
        """Returns the image histogram extracted from the image."""
        return self.image_obj.img_hist

    @property
    def image_2d(self) -> MatLike:
        """Returns OpenCV 2D version of the image (first slice/frame/image in the batch)."""
        return self.image_obj.img_2d

    @property
    def image_3d(self) -> list[MatLike]:
        """Returns the 3D version of the image as a list of OpenCV arrays."""
        images = [obj.img_2d for obj in self.image_obj_3d]
        return images

    @property
    def binary_image_2d(self) -> MatLike:
        """Returns OpenCV version of the binary image (first slice/frame/image in the batch)."""
        # img_bin_rgb = cv2.cvtColor(self.image_obj.img_bin, cv2.COLOR_BGR2RGB)
        return self.image_obj.img_bin

    @property
    def binary_image_3d(self) -> list[MatLike]:
        """Returns the 3D version of the binary image as a list of OpenCV arrays."""
        bin_images = [obj.img_bin for obj in self.image_obj_3d]
        return bin_images

    @property
    def processed_image_2d(self) -> MatLike:
        """Returns the modified OpenCV version of the image (first slice/frame/image in the batch)."""
        return self.image_obj.img_mod

    @property
    def processed_image_3d(self) -> list[MatLike]:
        """Returns the 3D version of the modified image as a list of OpenCV arrays."""
        mod_images = [obj.img_mod for obj in self.image_obj_3d]
        return mod_images

    @property
    def graph_image(self) -> MatLike:
        """Returns OpenCV version of the extracted graph image (first slice/frame/image in the batch)."""
        return self.graph_obj.img_ntwk

    def _load_img_from_file(self, file: list | str):
        """
        Read the image and save it as an OpenCV object.

        Most 3D images are like layers of multiple image frames layered on-top of each other. The image frames may be
        images of the same object/item through time or through space (i.e., from different angles). Our approach is to
        separate these frames, extract GT graphs from them, and then the layer back from the extracted graphs in the same order.

        Our software will display all the frames retrieved from the 3D image (automatically downsample large ones
        depending on the user-selected re-scaling options), and allows the user to select which frames to run
        GT computations on. (Some frames are just too noisy to be used.)

        Again, our software provides a button that allows the user to select which frames are used to reconstruct the
        layered GT graphs in the same order as their respective frames.

        :param file: The file path.
        :return: list[ImageProcessor.ImageBatch]
        """

        # First file if it's a list
        ext = os.path.splitext(file[0])[1].lower() if (type(file) is list) else os.path.splitext(file)[1].lower()
        try:
            if ext in ['.png', '.jpg', '.jpeg', '.bmp']:
                image_groups = defaultdict(list)
                if type(file) is list:
                    for img in file:
                        # Create clusters/groups of similar size images
                        frame = cv2.imread(img, cv2.IMREAD_UNCHANGED)
                        h, w = frame.shape[:2]
                        image_groups[(h, w)].append(frame)
                else:
                    # Load standard 2D images with OpenCV
                    image = cv2.imread(file, cv2.IMREAD_UNCHANGED)
                    if image is None:
                        raise ValueError(f"Failed to load {file}")
                    # Cluster the images into batches based on (h, w) size
                    h, w = image.shape[:2]
                    image_groups[(h, w)].append(image)
                img_batch_groups = ImageProcessor.create_img_batch_groups(image_groups, self._config_file,
                                                                          self._auto_scale)
                return img_batch_groups
            elif ext in ['.tif', '.tiff', '.qptiff']:
                image_groups = defaultdict(list)
                if type(file) is list:
                    for img in file:
                        # Create clusters/groups of similar size images
                        frame = cv2.imread(img, cv2.IMREAD_UNCHANGED)
                        h, w = frame.shape[:2]
                        image_groups[(h, w)].append(frame)
                else:
                    # Try load multi-page TIFF using PIL
                    img = Image.open(file)
                    while True:
                        # Create clusters/groups of similar size images
                        frame = np.array(img)  # Convert the current frame to the numpy array
                        # Cluster the images into batches based on (h, w) size
                        h, w = frame.shape[:2]
                        image_groups[(h, w)].append(frame)
                        try:
                            # Move to the next frame
                            img.seek(img.tell() + 1)
                        except EOFError:
                            # Stop when all frames are read
                            break
                img_batch_groups = ImageProcessor.create_img_batch_groups(image_groups, self._config_file,
                                                                          self._auto_scale)
                return img_batch_groups
            elif ext in ['.nii', '.nii.gz']:
                """# Load NIfTI image using nibabel
                img_nib = nib.load(file)
                data = img_nib.get_fdata()
                # Normalize and convert to uint8 for OpenCV compatibility
                data = cv2.normalize(data, None, 0, 255, cv2.NORM_MINMAX).astype(np.uint8)
                return data"""
                return []
            elif ext == '.dcm':
                """# Load DICOM image using pydicom
                dcm = pydicom.dcmread(file)
                data = dcm.pixel_array
                # Normalize and convert to uint8 if needed
                data = cv2.normalize(data, None, 0, 255, cv2.NORM_MINMAX).astype(np.uint8)
                return data"""
                return []
            else:
                raise ValueError(f"Unsupported file format: {ext}")
        except Exception as err:
            logging.exception(f"Error loading {file}:", err, extra={'user': 'SGT Logs'})
            # self.update_status([-1, f"Failed to load {file}: {err}"])
            return None

    def _initialize_image_batches(self, img_batches: list[ImageBatch]):
        """
        Retrieve all image slices of the selected image batch. If the image is 2D, only one slice exists
        if it is 3D, then multiple slices exist.
        """

        # Check if image batches exist
        if len(img_batches) == 0:
            raise ValueError("No images available! Please add at least one image.")

        for i, img_batch in enumerate(img_batches):
            img_data = img_batch.numpy_image
            scale_factor = img_batch.scale_factor

            # Load images for processing
            if img_data is None:
                raise ValueError(f"Problem with images in batch {i}!")

            _, fmt_2d = BaseImage.check_alpha_channel(img_data)
            image_list = []
            if (len(img_data.shape) >= 3) and (fmt_2d is None):
                # If the image has shape (d, h, w) and does not an alpha channel which is less than 4 - (h, w, a)
                image_list = [BaseImage(img, self._config_file, scale_factor) for img in img_data]
            else:
                img_obj = BaseImage(img_data, self._config_file, scale_factor)
                image_list.append(img_obj)

            is_2d = True
            if len(image_list) == 1:
                if len(image_list[0].img_2d.shape) == 3 and image_list[0].has_alpha_channel:
                    logging.info("Image is 2D with Alpha Channel.", extra={'user': 'SGT Logs'})
                    # self.update_status([101, "Image is 2D with Alpha Channel"])
                else:
                    logging.info("Image is 2D.", extra={'user': 'SGT Logs'})
                    # self.update_status([101, "Image is 2D"])
            elif len(image_list) > 1:
                is_2d = False
                logging.info("Image is 3D.", extra={'user': 'SGT Logs'})
                # self.update_status([101, "Image is 3D"])

            img_batch.images = image_list
            img_batch.is_2d = is_2d
            self.update_image_props(img_batch)
        self._image_batches = img_batches

    def select_image_batch(self, sel_batch_idx: int, selected_images: set = None):
        """
        Update the selected image batch and the selected image slices.

        Args:
            sel_batch_idx: index of the selected image batch
            selected_images: indices of the selected image slices.

        Returns:

        """

        if sel_batch_idx >= len(self._image_batches):
            raise ValueError(
                f"Selected image batch {sel_batch_idx} out of range! Select in range 0-{len(self._image_batches)}")

        self._selected_batch_index = sel_batch_idx
        self.update_image_props(self._image_batches[sel_batch_idx])
        self.reset_img_filters()

        if selected_images is None:
            return

        if type(selected_images) is set:
            self._image_batches[sel_batch_idx].selected_images_idx = selected_images

    def track_progress(self, value, msg):
        self.update_status([value, msg])

    def apply_img_filters(self, filter_type=2):
        """
        Executes function for processing image filters and converting the resulting image into a binary.

        Filter Types:
        1 - Just Image Filters
        2 - Both Image and Binary (1 and 2) Filters

        :return: None
        """

        if self.selected_batch.is_graph_only:
            return

        self.update_status([10, "Processing image..."])
        if filter_type == 2:
            self.reset_img_filters()

        sel_batch = self.selected_batch
        progress = 10
        incr = 90 / len(sel_batch.images) - 1
        for i in range(len(sel_batch.images)):
            img_obj = sel_batch.images[i]
            if i not in sel_batch.selected_images_idx:
                img_obj.img_mod, img_obj.img_bin = None, None
                continue

            if progress < 100:
                progress += incr
                self.update_status([progress, "Image processing in progress..."])

            img_data = img_obj.img_2d.copy()
            img_obj.img_mod = img_obj.process_img(image=img_data)

            if filter_type == 2:
                img_mod = img_obj.img_mod.copy()
                img_obj.img_bin = img_obj.binarize_img(img_mod)
                img_obj.img_mod = img_mod
            img_obj.get_pixel_width()
        self.update_status([100, "Image processing complete..."])

    def reset_img_filters(self):
        """Delete existing filters that have been applied on the image."""
        sel_batch = self.selected_batch
        for img_obj in sel_batch.images:
            img_obj.img_mod, img_obj.img_bin = None, None
            sel_batch.graph_obj.reset_graph()

    def apply_img_scaling(self):
        """Re-scale (downsample or up-sample) a 2D image or 3D images to a specified size"""

        # scale_factor = 1
        sel_batch = self.selected_batch
        if len(sel_batch.images) <= 0:
            return

        scale_size = 0
        for scale_item in sel_batch.scaling_options:
            try:
                scale_size = scale_item["dataValue"] if scale_item["value"] == 1 else scale_size
            except KeyError:
                continue

        if scale_size <= 0:
            return

        img_px_size = 1
        for img_obj in sel_batch.images:
            img = img_obj.img_raw
            temp_px = max(img.shape[0], img.shape[1])
            img_px_size = temp_px if temp_px > img_px_size else img_px_size
        scale_factor = scale_size / img_px_size

        # Resize (Downsample) all frames to the smaller pixel size while maintaining the aspect ratio
        for img_obj in sel_batch.images:
            img = img_obj.img_raw.copy()
            scale_size = scale_factor * max(img.shape[0], img.shape[1])
            img_small, _ = BaseImage.resize_img(scale_size, img)
            if img_small is None:
                # raise Exception("Unable to Rescale Image")
                return
            img_obj.img_2d = img_small
            img_obj.scale_factor = scale_factor
        self.update_image_props(sel_batch)

    def crop_image(self, x: int, y: int, crop_w: int, crop_h: int, actual_w: int, actual_h: int):
        """
        A function that crops images into a new box dimension.

        :param x: Left coordinate of cropping box.
        :param y: Top coordinate of cropping box.
        :param crop_w: Width of cropping box.
        :param crop_h: Height of cropping box.
        :param actual_w: Width of actual image.
        :param actual_h: Height of actual image.
        """
        sel_batch = self.selected_batch
        if len(sel_batch.selected_images_idx) > 0:
            [sel_batch.images[i].apply_img_crop(x, y, crop_w, crop_h, actual_w, actual_h) for i in
             sel_batch.selected_images_idx]
        self.update_image_props(sel_batch)
        self.selected_batch_view = 'processed'

    def undo_cropping(self):
        """
        A function that restores the image to its original size.
        """
        sel_batch = self.selected_batch
        if len(sel_batch.selected_images_idx) > 0:
            [sel_batch.images[i].init_image() for i in sel_batch.selected_images_idx]
        self.update_image_props(sel_batch)

    def metaheuristic_image_configs(self) -> dict | None:
        """
        A function that runs metaheuristic algorithms (Genetic Algorithm and Hill-climbing Algorithm) to find the best
        image configurations for extracting accurate graphs from SEM images.

        :return: A dictionary containing the best candidate's image configuration settings.
        """

        # TO BE DELETED
        def _print_configs(title: str = ""):
            print(f"{title}")
            print(
                f"Configs: {filter_space.best_candidate.img_configs}\n"
                f"Cost: {filter_space.best_candidate.std_cost}\n"
                f"Graph Accuracy: {filter_space.best_candidate.graph_accuracy}\n"
                f"")

        self.update_status([0, "Starting image-filter search..."])
        opt_model = self._configs
        opt_model["find_filter_selections"]["value"] = 1    # TO BE DELETED
        opt_model["find_filter_values"]["value"] = 1        # TO BE DELETED
        opt_model["find_brightness_contrast"]["value"] = 1  # TO BE DELETED
        max_iters = opt_model["max_iterations"]["value"]
        ga_init_pop = opt_model["genetic_alg_initial_pop"]["value"]

        if self.abort:
            self.update_status([-1, "Task aborted!"])
            return None

        # 1. Create a search space
        self.update_status([20, "Creating search space..."])
        filter_space = FilterSearchSpace.build_search_space(self.image_obj, initial_pop=ga_init_pop)
        _print_configs("Default Configs")

        if opt_model["find_filter_selections"]["value"] == 1:
            # 2. Run the Hill-climbing algorithm to find the best "image config combination"
            self.update_status([50, "Searching for best image filter selections..."])
            try:
                # filter_space.ignore_candidates.append(filter_space.best_candidate.position)
                sgt_hill_climbing_algorithm(filter_space, self.image_obj, max_iters=max_iters)
                _print_configs("Selected Configs")
            except AbortException as err:
                self.abort = True
                logging.exception(f"Error finding best apply selections:", err, extra={'user': 'SGT Logs'})
                self.update_status([-1, f"Find Filter Selections Error: {err}"])
                return None

        if opt_model["find_filter_values"]["value"] == 1:
            # 3. Run the Genetic Algorithm to find the best "image filter values"
            self.update_status([65, "Searching for best image filter values..."])
            try:
                val_search_space = filter_space.best_candidate.value_space
                val_img_configs = sgt_genetic_algorithm(val_search_space, self.image_obj, generations=max_iters,
                                                        pop_size=ga_init_pop)
                filter_space.best_candidate.std_cost = val_search_space.best_candidate.std_cost
                filter_space.img_configs = val_img_configs
                _print_configs("Best Image Configs")
            except AbortException as err:
                self.abort = True
                logging.exception(f"Error best filter values:", err, extra={'user': 'SGT Logs'})
                self.update_status([-1, f"Find Filter Values Error: {err}"])
                return None

        if opt_model["find_brightness_contrast"]["value"] == 1:
            # 4. Run the Genetic Algorithm to find the best "brightness/contrast values" (only if 'val_search_space' fxn fails)
            self.update_status([80, "Searching for best image filter values..."])
            try:
                bright_search_space = filter_space.best_candidate.brightness_space
                brt_img_configs = sgt_genetic_algorithm(bright_search_space, self.image_obj, generations=max_iters,
                                                        pop_size=ga_init_pop)
                filter_space.best_candidate.std_cost = bright_search_space.best_candidate.std_cost
                filter_space.img_configs = brt_img_configs
                _print_configs("Best Brightness/Contrast Configs")
            except AbortException as err:
                self.abort = True
                logging.exception(f"Error best brightness/contrast values:", err, extra={'user': 'SGT Logs'})
                self.update_status([-1, f"Find Brightness/Contrast Error: {err}"])
                return None
        return filter_space.best_candidate.img_configs

    def build_graph_network(self):
        """Generates or extracts graphs of selected images."""

        self.update_status([0, "Starting graph extraction..."])
        try:
            # Get the selected batch
            self.selected_batch_view= 'graph'
            sel_batch = self.selected_batch
            sel_images = self.get_batch_images(sel_batch)

            if sel_batch.is_graph_only:
                self.update_status([20, "Fetching graph file..."])
                f_name, out_dir = self.get_filenames(file_path=self._graph_file)

                sel_batch.graph_obj.abort = False
                sel_batch.graph_obj.add_listener(self.track_progress)
                sel_batch.graph_obj.fit_graph(out_dir, input_data=self._graph_file, file_name=f_name)
                sel_batch.graph_obj.remove_listener(self.track_progress)
            else:
                # Get binary image
                self.update_status([20, "Getting binary image..."])
                img_bin = [img.img_bin for img in sel_images]
                img_bin = np.asarray(img_bin)

                # Check if filters have been applied
                if img_bin[0] is None:
                    self.update_status([101, "No filters applied! Please wait, applying image filters."])
                    self.apply_img_filters()
                    self.build_graph_network()
                    return

                # Get the selected batch's graph object and generate the graph
                px_size = float(sel_batch.images[0].configs["pixel_width"]["value"])  # First BaseImage in batch
                rho_val = float(sel_batch.images[0].configs["resistivity"]["value"])  # First BaseImage in batch
                f_name, out_dir = self.get_filenames()

                sel_batch.graph_obj.abort = False
                sel_batch.graph_obj.add_listener(self.track_progress)
                sel_batch.graph_obj.fit_graph(out_dir, img_bin, sel_batch.is_2d, px_size, rho_val, file_name=f_name)

            self.update_status([95, "Plotting graph network..."])
            self.draw_graph_image(sel_batch)

            sel_batch.graph_obj.remove_listener(self.track_progress)
            self.abort = sel_batch.graph_obj.abort
            if self.abort:
                self.selected_batch_view = 'processed'
                return
        except Exception as err:
            self.abort = True
            logging.exception("Graph Extraction Error: %s", err, extra={'user': 'SGT Logs'})
            self.update_status([-1, f"Graph Extraction Error: {err}"])
            return

    def build_graph_from_patches(self, num_kernels: int, patch_count_per_kernel: int, img_padding: tuple = (0, 0),
                                 compute_avg: bool = False):
        """
        Extracts graphs from smaller square patches of selected images.

        Given `num_square_filters` (k), the method generates k square filters/windows, each of sizes NxN—where N is
        a distinct value computed or estimated for each filter.

        For every NxN window, it randomly selects `patch_count_per_filter` (m) patches (aligned with the window)
        from across the entire image.

        :param num_kernels: Number of square kernels/filters to generate.
        :param patch_count_per_kernel: Number of patches per filter.
        :param img_padding: Padding around the image.
        :param compute_avg: If True, allows for computing of GT params from 95% of the original image sampled at different
        locations (by extracting the graphs at these locations).

        """
        # Get the selected batch
        patch_count_per_kernel = patch_count_per_kernel if patch_count_per_kernel > 5 else 5
        graph_configs = self.graph_obj.configs
        img_obj = self.image_obj  # ONLY works for 2D

        def extract_cropped_image_patches() -> list[MatLike]:
            """A method that extracts 4 filters from the original binary image. Each filter is the of size approximately
            90% of the original image height and width. This method ensures exactly four patches are extracted from
            the corners. Compute GT descriptors of 90% original image at different locations (to get their averages)
            """
            # Create a kernel that is 95% the size of the image
            img_bin = self.binary_image_2d
            h, w = img_bin.shape
            k_h, k_w = int(0.95 * h), int(0.95 * w)

            # Coordinates for the 4 positions (top-left, top-right, bottom-left, bottom-right)
            slide_positions = [
                (0, 0),  # top-left
                (0, w - k_w),  # top-right
                (h - k_h, 0),  # bottom-left
                (h - k_h, w - k_w),  # bottom-right
            ]

            # Retrieve kernel patches and compute GT descriptors
            lst_img_90pct = []
            for y, x in slide_positions:
                img_90pct = img_bin[y:y + k_h, x:x + k_w]
                lst_img_90pct.append(img_90pct)
            return lst_img_90pct

        def retrieve_kernel_patches(img: MatLike, num_filters: int, num_patches: int, padding: tuple) -> list[BaseImage.ScalingKernel]:
            """
            Perform an incomplete convolution operation that breaks down an image into smaller square mini-images.
            Extract all patches from the image based on filter size, stride, and padding, similar to
            CNN convolution but without applying the multiplication and addition operations. The kernel patches
            are retrieved from random/deterministic locations in the image.

            :param img: OpenCV image.
            :param num_filters: Number of convolution kernels/filters.
            :param num_patches: Number of patches to extract per filter window size.
            :param padding: Padding value (pad_y, pad_x).
            :return: List of convolved images.
            """

            def estimate_kernel_size(parent_width, num) -> int:
                """
                Applies a non-linear function to compute the width-size of a filter based on its index location.
                :param parent_width: Width of parent image.
                :param num: Index of filter.
                """
                # return int(parent_width / ((2*num) + 4))
                # est_w = int((parent_width * np.exp(-0.3 * num) / 4))  # Exponential decay
                est_w = int((parent_width - 10) * (1 - (num / num_kernels)))
                return max(50, est_w)  # Avoid too small sizes

            def extract_random_patches(kernel_dim) -> list[MatLike]:
                """
                Retrieve kernel patches at random locations in the image.
                Args:
                    kernel_dim: dimension of kernel.

                Returns:
                    list of extracted patches each of size kernel_dim.
                """

                lst_patches = []
                img_h, img_w = img.shape[:2]
                k_h, k_w = kernel_dim, kernel_dim
                for _ in range(num_patches):
                    # Random top-left corner
                    x = np.random.randint(0, img_w - k_w)
                    y = np.random.randint(0, img_h - k_h)

                    patch = img_padded[y:y + k_h, x:x + k_w].copy()
                    lst_patches.append(patch)
                    # print(f"Filter Shape: {patch.shape} at strides: x={x}, y={y}")
                return lst_patches

            if img is None:
                return []

            # Initialize Parameters
            lst_img_filter = []

            # Pad the image
            pad_h, pad_w = padding
            img_padded = np.pad(img.copy(), ((pad_h, pad_h), (pad_w, pad_w)), mode='constant')

            # Get largest image dimension (height or width)
            h, w = img.shape[:2]
            max_dim = h if h < w else w

            # For each filter-window: (1) estimate HxW, (2) stride size and (3) sliding window patch retrieval
            for k in range(num_filters):
                # (1) Estimate HxW dimensions of filter
                kernel_size = estimate_kernel_size(max_dim, k)

                # (2) Retrieve multiple patches of size (k_h, k_w)
                lst_kernel_patches = extract_random_patches(kernel_size)

                # Save filter parameters in dict
                img_filter = BaseImage.ScalingKernel(
                    image_patches=lst_kernel_patches,
                    kernel_shape=(kernel_size, kernel_size),
                )
                lst_img_filter.append(img_filter)

                # Stop loop if filter size is too small
                if kernel_size <= 50:
                    break
            return lst_img_filter

        if len(img_obj.image_filters) <= 0:
            # Scaling patches (square kernel)
            lst_filters = retrieve_kernel_patches(img_obj.img_bin, num_kernels, patch_count_per_kernel, img_padding)

            # Average patches with sizes 90% of the image (rectangular kernel)
            if compute_avg:
                self.update_status([66, "Computing GT parameters on 95% of image at 4 locations..."])
                lst_img_filters = extract_cropped_image_patches()
                c_h, c_w = lst_img_filters[0].shape[:2]
                crop_filter = BaseImage.ScalingKernel(
                    image_patches=lst_img_filters,
                    kernel_shape=(c_h, c_w),
                )
                lst_filters.append(crop_filter)
            img_obj.image_filters = lst_filters

        filter_count = len(img_obj.image_filters)
        graph_groups = defaultdict(list)
        for i, scale_filter in enumerate(img_obj.image_filters):
            self.update_status([101, f"Extracting random graphs using image filter {i + 1}/{filter_count}..."])
            # num_img_patches = len(scale_filter.image_patches)
            for bin_img_patch in scale_filter.image_patches:
                graph_patch = FiberNetworkBuilder(cfg_file=self._config_file)
                graph_patch.configs = graph_configs
                nx_graph = graph_patch.extract_graph(bin_img_patch, is_img_2d=True)
                success = graph_patch.verify_graph(nx_graph)
                # success = graph_patch.extract_graph(bin_img_patch, is_img_2d=True)
                if success:
                    height, width = bin_img_patch.shape
                    graph_groups[(height, width)].append(graph_patch.nx_giant_graph)
                else:
                    self.update_status([-1, f"Filter {bin_img_patch.shape} graph extraction failed!"])
        return graph_groups

    def get_filenames(self, file_path: str = None):
        """
        Splits the image path into file name and image directory.

        :param file_path: Path to the file of interest.

        Returns:
            filename (str): image file name., output_dir (str): image directory path.
        """

        img_dir, filename = os.path.split(self._img_path) if file_path is None else os.path.split(file_path)
        output_dir = img_dir if self._output_dir == '' else self._output_dir

        for ext in ALLOWED_IMG_EXTENSIONS:
            ext = ext.replace('*', '')
            pattern = re.escape(ext) + r'$'
            filename = re.sub(pattern, '', filename)

        for ext in ALLOWED_GRAPH_FILE_EXTENSIONS:
            ext = ext.replace('*', '')
            pattern = re.escape(ext) + r'$'
            filename = re.sub(pattern, '', filename)

        for ext in ALLOWED_3D_IMG_EXTENSIONS:
            ext = ext.replace('*', '')
            pattern = re.escape(ext) + r'$'
            filename = re.sub(pattern, '', filename)
        return filename, output_dir

    def get_batch_images(self, selected_batch: ImageBatch):
        """
        Get indices of selected images.
        :param selected_batch: The selected batch ImageBatch object.
        """
        if selected_batch is None:
            selected_batch = self.selected_batch

        sel_images = [selected_batch.images[i] for i in selected_batch.selected_images_idx]
        return sel_images

    def update_image_props(self, selected_batch: ImageBatch = None):
        """
        A method that retrieves image properties and stores them in a list-array.

        :param selected_batch: ImageBatch data object.

        Returns: list of image properties

        """

        if selected_batch is None:
            return

        f_name, _ = self.get_filenames()
        if len(selected_batch.images) > 1:
            # (Depth, Height, Width, Channels)
            alpha_channel = selected_batch.images[0].has_alpha_channel  # first image
            fmt = "Multi + Alpha" if alpha_channel else "Multi"
            num_dim = 3
        elif len(selected_batch.images) == 1:
            if selected_batch.is_graph_only:
                return
            else:
                # (Height, Width, Channels)
                _, fmt = BaseImage.check_alpha_channel(selected_batch.images[0].img_raw)  # first image
                num_dim = 2
        else:
            # No Image Found
            return

        slices = 0
        height, width = selected_batch.images[0].img_2d.shape[:2]  # first image
        if num_dim >= 3:
            slices = len(selected_batch.images)

        props = [
            ["Name", f_name],
            ["Height x Width", f"({height} x {width}) pixels"] if slices == 0
            else ["Depth x H x W", f"({slices} x {height} x {width}) pixels"],
            ["Dimensions", f"{num_dim}D"],
            ["Format", f"{fmt}"],
            # ["Pixel Size", "2nm x 2nm"]
        ]
        selected_batch.props = props

    def save_images_to_file(self):
        """
        Write images to a file.
        """

        sel_batch = self.selected_batch
        sel_images = self.get_batch_images(sel_batch)
        is_3d = True if len(sel_images) > 1 else False
        img_file_name, out_dir = self.get_filenames()

        for i, img in enumerate(sel_images):
            if img.configs["save_images"]["value"] == 0:
                continue

            filename = f"{img_file_name}_Frame{i}" if is_3d else ''
            pr_filename = filename + "_processed.jpg"
            bin_filename = filename + "_binary.jpg"
            img_file = os.path.join(out_dir, pr_filename)
            bin_file = os.path.join(out_dir, bin_filename)

            if img.img_mod is not None:
                cv2.imwrite(str(img_file), img.img_mod)

            if img.img_bin is not None:
                cv2.imwrite(str(bin_file), img.img_bin)

    def draw_graph_image(self, sel_batch: ImageBatch, show_giant_only: bool = False):
        """
        Use Matplotlib to draw the extracted graph which is superimposed on the processed image.

        :param sel_batch: ImageBatch data object.
        :param show_giant_only: If True, only draw the largest/giant graph on the processed image.
        """
        sel_images = self.get_batch_images(sel_batch)

        if len(sel_images) > 0:
            img_3d = [img.img_2d for img in sel_images]
            img_3d = np.asarray(img_3d)
        else:
            img_3d = None

        if sel_batch.graph_obj is None:
            return

        plt_fig = sel_batch.graph_obj.plot_graph_network(image_arr=img_3d, giant_only=show_giant_only)
        if plt_fig is not None:
            self.update_status([98, "Graph image created."])
            sel_batch.graph_obj.img_ntwk = plot_to_opencv(plt_fig)

    # MODIFIED TO EXCLUDE 3D IMAGES (TO BE REVISITED LATER)
    # Problems:
    # 1. Merge Nodes
    # 2. Prune dangling edges
    # 3. Matplotlib plot nodes and edges
    @staticmethod
    def create_img_batch_groups(img_groups: defaultdict, cfg_file: str, auto_scale: bool):
        """"""

        def get_scaling_options(orig_size: float):
            """"""
            orig_size = int(orig_size)
            if orig_size > 2048:
                recommended_size = 1024
                scaling_options = [1024, 2048, int(orig_size * 0.25), int(orig_size * 0.5), int(orig_size * 0.75),
                                   orig_size]
            elif orig_size > 1024:
                recommended_size = 1024
                scaling_options = [1024, int(orig_size * 0.25), int(orig_size * 0.5), int(orig_size * 0.75), orig_size]
            else:
                recommended_size = orig_size
                scaling_options = [int(orig_size * 0.25), int(orig_size * 0.5), int(orig_size * 0.75), orig_size]

            # Remove duplicates and arrange in ascending order
            scaling_options = sorted(list(set(scaling_options)))
            scaling_data = []
            for val in scaling_options:
                data = {"text": f"{val} px", "value": 0, "dataValue": val}
                if val == orig_size:
                    data["text"] = f"{data['text']}*"

                if val == recommended_size:
                    data["text"] = f"{data['text']} (recommended)"
                    data["value"] = 1 if auto_scale else 0
                scaling_data.append(data)
            return scaling_data

        def rescale_img(image_data, scale_options):
            """Downsample or up-sample image to a specified pixel size."""

            scale_factor = 1
            img_2d, img_3d = None, None

            if image_data is None:
                return None, scale_factor

            scale_size = 0
            for scale_item in scale_options:
                try:
                    scale_size = scale_item["dataValue"] if scale_item["value"] == 1 else scale_size
                except KeyError:
                    continue

            if scale_size <= 0:
                return None, scale_factor

            # if type(image_data) is np.ndarray:
            has_alpha, fmt_2d = BaseImage.check_alpha_channel(image_data)
            if (len(image_data.shape) == 2) or has_alpha:
                # If the image has shape (h, w) or shape (h, w, a), where 'a' - alpha channel which is less than 4
                img_2d, scale_factor = BaseImage.resize_img(scale_size, image_data)
                return img_2d, scale_factor

            # if type(image_data) is list:
            if (len(image_data.shape) >= 3) and (fmt_2d is None):
                # If the image has shape (d, h, w) and third is not alpha channel
                img_3d = []
                for img in image_data:
                    img_small, scale_factor = BaseImage.resize_img(scale_size, img)
                    img_3d.append(img_small)
            return np.array(img_3d), scale_factor

        img_info_list = []
        for (h, w), images in img_groups.items():
            images_small = []
            scaling_factor = 1
            scaling_opts = []
            images = np.array(images)
            max_size = max(h, w)
            if max_size > 0 and auto_scale:
                scaling_opts = get_scaling_options(max_size)
                images_small, scaling_factor = rescale_img(images, scaling_opts)

            # Convert back to numpy arrays
            images = images_small if len(images_small) > 0 else images
            images = np.array([images[0]])  # REMOVE TO ALLOW 3D
            views  = [
                {"text": "Original Image", "dataValue": "original", "value": 1, "visible": 1 },
                {"text": "Binary Image", "dataValue": "binary", "value": 0, "visible": 1 },
                {"text": "Processed Image", "dataValue": "processed", "value": 0, "visible": 1 },
                {"text": "Extracted Graph", "dataValue": "graph", "value": 0, "visible": 1}
            ]
            img_batch = ImageProcessor.ImageBatch(
                numpy_image=images,
                images=[],
                graph_obj=FiberNetworkBuilder(cfg_file=cfg_file),
                shape=(h, w),
                props=[],
                is_2d=True,
                is_graph_only=False,
                scale_factor=scaling_factor,
                scaling_options=scaling_opts,
                selected_images_idx=set(range(len(images))),
                view_options=views,
            )
            img_info_list.append(img_batch)
            break  # REMOVE TO ALLOW 3D
        return img_info_list

    @classmethod
    def from_image_file(cls, img_path: str, out_folder: str = "", config_file: str = "", allow_auto_scale: bool = True) -> tuple["ImageProcessor", str]:
        """
        Creates an ImageProcessor object. Make sure the image path exists, is verified, and points to an image.
        :param img_path: Path to the image to be processed
        :param out_folder: Path to the output directory
        :param config_file: Path to the config file
        :param allow_auto_scale: Allows automatic scaling of the image
        :return: ImageProcessor object.
        """

        # Get the image path and folder
        img_files = []
        img_dir, img_file = os.path.split(str(img_path))
        img_file_ext = os.path.splitext(img_file)[1].lower()

        is_prefix = True
        # Regex pattern to extract the prefix (non-digit characters at the beginning of the file name)
        img_name_pattern = re.match(r'^([a-zA-Z_]+)(\d+)(?=\.[a-zA-Z]+$)', img_file)
        if img_name_pattern is None:
            # Regex pattern to extract the suffix (non-digit characters at the end of the file name)
            is_prefix = False
            img_name_pattern = re.match(r'^\d+([a-zA-Z_]+)(?=\.[a-zA-Z]+$)', img_file)

        # If 3D file (ignore multiple input files)
        """allowed_3d_extensions = tuple(ext[1:] if ext.startswith('*.') else ext for ext in ALLOWED_3D_IMG_EXTENSIONS)
        if img_path.endswith(allowed_3d_extensions):
            img_name_pattern = None
        """

        if img_name_pattern:
            img_files.append(img_path)
            f_name = img_name_pattern.group(1)
            name_pattern = re.compile(rf'^{f_name}\d+{re.escape(img_file_ext)}$', re.IGNORECASE) \
                if is_prefix else re.compile(rf'^\d+{f_name}{re.escape(img_file_ext)}$', re.IGNORECASE)

            # Check if 3D image slices exist in the image folder. Same file name but different number
            files = sorted(os.listdir(img_dir))
            for a_file in files:
                if a_file.endswith(img_file_ext):
                    if name_pattern.match(a_file):
                        img_files.append(os.path.join(img_dir, a_file))

        # Create the Output folder if it does not exist
        if out_folder != "":
            default_out_dir = out_folder
        else:
            out_dir_name = "sgt_files"
            default_out_dir = os.path.join(img_dir, out_dir_name)
        out_dir = os.path.normpath(default_out_dir)
        os.makedirs(out_dir, exist_ok=True)

        # Create the StructuralGT object
        input_file = img_files if len(img_files) > 1 else str(img_path)
        imp_obj = cls(input_file, out_dir, config_file, auto_scale=allow_auto_scale)
        imp_obj._initialize_image_batches(imp_obj._load_img_from_file(img_path))
        return imp_obj, img_file

    @classmethod
    def from_graph_file(cls, file_path: str, out_folder: str = "") -> tuple["ImageProcessor", str]:
        """
        Creates an ImageProcessor object. Make sure the graph file path exists, is verified, and points to a
        CSV file or GSD/HOOMD file.
        :param file_path: Path to the graph file
        :param out_folder: Path to the output directory
        :return: ImageProcessor object.
        """
        # Separate graph path and folder
        file_dir, graph_file = os.path.split(str(file_path))
        file_ext = os.path.splitext(graph_file)[1].lower()
        allowed_extensions = tuple(ext[1:] if ext.startswith('*.') else ext for ext in ALLOWED_GRAPH_FILE_EXTENSIONS)
        if not graph_file.endswith(allowed_extensions):
            raise ValueError(f"Unsupported file format: {file_ext}")

        # Create the Output folder if it does not exist
        if out_folder != "":
            default_out_dir = out_folder
        else:
            out_dir_name = "sgt_files"
            default_out_dir = os.path.join(file_dir, out_dir_name)
        out_dir = os.path.normpath(default_out_dir)
        os.makedirs(out_dir, exist_ok=True)

        # Create the StructuralGT object
        imp_obj = cls("", out_dir, graph_file=str(file_path))

        # Create Graph Object from the added file_path
        graph_obj = FiberNetworkBuilder(cfg_file="")

        # Create an Image Batch data object and add it to the ImageProcessor object
        views = [
            {"text": "Original Image", "dataValue": "original", "value": 0, "visible": 0},
            {"text": "Binary Image", "dataValue": "binary", "value": 0, "visible": 0},
            {"text": "Processed Image", "dataValue": "processed", "value": 0, "visible": 0},
            {"text": "Extracted Graph", "dataValue": "graph", "value": 1, "visible": 1}
        ]
        img_batch = ImageProcessor.ImageBatch(
            numpy_image=np.array([None]),
            images=[BaseImage(None)],
            graph_obj=graph_obj,
            shape=(0,0),
            props=[],
            is_2d=True,
            is_graph_only=True,
            scale_factor=0.0,
            scaling_options=[],
            selected_images_idx=set(),
            view_options=views,
        )
        imp_obj._image_batches = [img_batch]
        return imp_obj, graph_file



from PySide6.QtCore import Qt, QAbstractItemModel, QModelIndex


class TreeItem:
    """ Represents a single item in the tree. """
    def __init__(self, data, parent=None):
        self._data = data  # Dictionary of data (id, text, value)
        self._parent = parent
        self.children = []

    def append_child(self, child):
        """ Adds a child node to this item. """
        self.children.append(child)

    def child(self, row):
        """ Returns the child at the specified row. """
        return self.children[row] if row < len(self.children) else None

    def child_count(self):
        """ Returns the number of children. """
        return len(self.children)

    def column_count(self):
        """ Returns the number of columns (3: id, text, value). """
        return len(self._data)

    def data(self, column):
        """ Returns the data for a specific column. """
        if column == 0:
            return self._data.get("id", "")
        elif column == 1:
            return self._data.get("text", "")
        elif column == 2:
            return self._data.get("value", 0)
        return None

    def set_data(self, column, value):
        """ Sets the data for a specific column. """
        if column == 0:
            self._data["id"] = value
        elif column == 1:
            self._data["text"] = value
        elif column == 2:
            self._data["value"] = value
        return True

    def parent(self):
        """ Returns the parent of this item. """
        return self._parent

    def row(self):
        """ Returns this item's index in its parent's children list. """
        if self._parent:
            return self._parent.children.index(self)
        return 0


class TreeModel(QAbstractItemModel):
    """ QAbstractItemModel for displaying hierarchical data in QML. """
    # Qt.UserRole
    IdRole = Qt.ItemDataRole.UserRole + 1
    TextRole = Qt.ItemDataRole.UserRole + 2
    ValueRole = Qt.ItemDataRole.UserRole + 3

    def __init__(self, data, parent=None):
        super().__init__(parent)
        self._rootItem = TreeItem({"id": "root", "text": "Root", "value": None})
        self.setup_model_data(data, self._rootItem)
        # self.

    def setup_model_data(self, data_list, parent):
        """ Populates the model with data from a list of dictionaries. """
        for item_data in data_list:
            item = TreeItem(item_data, parent)
            parent.append_child(item)
            if 'items' in item_data:
                self.setup_model_data(item_data['items'], item)

    def rowCount(self, parent=QModelIndex()):
        """ Returns the children count of the given parent. """
        if not parent.isValid():
            return self._rootItem.child_count()
        return parent.internalPointer().child_count()

    def columnCount(self, parent=QModelIndex()):
        """ Returns the number of columns (fixed at 3). """
        return 1

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        """ Returns the data to be displayed. """
        if not index.isValid():
            return None
        item = index.internalPointer()
        # if role == Qt.DisplayRole:
        #    return item.data(index.column())
        if role == self.IdRole:
            return item.data(0)  # id
        elif role == self.TextRole:
            return item.data(1)  # text
        elif role == self.ValueRole:
            return item.data(2)  # value
        return None

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        if not index.isValid():
            return False
        item = index.internalPointer()
        col = 2  # for valueRole
        if item.set_data(col, value):
            self.dataChanged.emit(index, index, [role])
            return True
        return False

    def reset_data(self, new_data):
        """ Resets the data to be displayed. """
        self.beginResetModel()
        self._rootItem = TreeItem({"id": "root", "text": "Root", "value": None})
        self.setup_model_data(new_data, self._rootItem)
        self.endResetModel()
        self.dataChanged.emit(self.index(0,0), self.index(len(new_data), 0),
                              [self.IdRole, self.TextRole, self.ValueRole])

    def parent(self, index=QModelIndex()):
        """ Returns the parent index of a given child index. """
        if not index.isValid():
            return QModelIndex()

        item = index.internalPointer()
        parent_item = item.parent()

        if parent_item == self._rootItem:
            return QModelIndex()

        return self.createIndex(parent_item.row(), 0, parent_item)

    def index(self, row, column, parent=QModelIndex()):
        """ Returns the index of a given row and column. """
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        parent_item = self._rootItem if not parent.isValid() else parent.internalPointer()
        child_item = parent_item.child(row)

        if child_item:
            return self.createIndex(row, column, child_item)

        return QModelIndex()

    def roleNames(self):
        return {
            self.IdRole: b"id",
            self.TextRole: b"text",
            self.ValueRole: b"value",
        }



# SPDX-License-Identifier: GNU GPL v3

"""
Pyside6 implementation of StructuralGT user interface.
"""

import os
import sys
from PySide6.QtCore import QObject
from PySide6.QtWidgets import QApplication
from PySide6.QtQml import QQmlApplicationEngine

from .gui_mcw.controller import MainController
from .gui_mcw.image_provider import ImageProvider


class PySideApp(QObject):

    def __init__(self):
        super().__init__()
        self.app = QApplication(sys.argv)
        self._ui_engine = QQmlApplicationEngine()
        # Register Controller for Dynamic Updates
        self._controller = MainController(qml_app=self.app)
        # Register Image Provider
        self._image_provider = ImageProvider(self._controller)
        self._qml_file = 'sgt_qml/MainWindow.qml'

        # Set Models in QML Context
        self._ui_engine.rootContext().setContextProperty("imgThumbnailModel", self._controller.imgThumbnailModel)
        self._ui_engine.rootContext().setContextProperty("imagePropsModel", self._controller.imagePropsModel)
        self._ui_engine.rootContext().setContextProperty("graphPropsModel", self._controller.graphPropsModel)
        self._ui_engine.rootContext().setContextProperty("graphComputeModel", self._controller.graphComputeModel)
        self._ui_engine.rootContext().setContextProperty("microscopyPropsModel", self._controller.microscopyPropsModel)

        self._ui_engine.rootContext().setContextProperty("gteTreeModel", self._controller.gteTreeModel)
        self._ui_engine.rootContext().setContextProperty("gtcListModel", self._controller.gtcListModel)
        self._ui_engine.rootContext().setContextProperty("gtcScalingModel", self._controller.gtcScalingModel)
        self._ui_engine.rootContext().setContextProperty("exportGraphModel", self._controller.exportGraphModel)
        self._ui_engine.rootContext().setContextProperty("imgBatchModel", self._controller.imgBatchModel)
        self._ui_engine.rootContext().setContextProperty("imgControlModel", self._controller.imgControlModel)
        self._ui_engine.rootContext().setContextProperty("imgBinFilterModel", self._controller.imgBinFilterModel)
        self._ui_engine.rootContext().setContextProperty("imgFilterModel", self._controller.imgFilterModel)
        self._ui_engine.rootContext().setContextProperty("imgScaleOptionModel", self._controller.imgScaleOptionModel)
        self._ui_engine.rootContext().setContextProperty("imgViewOptionModel", self._controller.imgViewOptionModel)
        self._ui_engine.rootContext().setContextProperty("saveImgModel", self._controller.saveImgModel)
        self._ui_engine.rootContext().setContextProperty("img3dGridModel", self._controller.img3dGridModel)
        self._ui_engine.rootContext().setContextProperty("imgHistogramModel", self._controller.imgHistogramModel)
        self._ui_engine.rootContext().setContextProperty("mainController", self._controller)
        self._ui_engine.addImageProvider("imageProvider", self._image_provider)

        # Load UI
        # Get the directory of the current script
        qml_dir = os.path.dirname(os.path.abspath(__file__))
        qml_path = os.path.join(qml_dir, self._qml_file)
        self._ui_engine.load(qml_path)
        if not self._ui_engine.rootObjects():
            sys.exit(-1)

    @classmethod
    def start(cls) -> None:
        """
        Initialize and run the PySide GUI application.
        """
        gui_app = cls()
        sys.exit(gui_app.app.exec())



# SPDX-License-Identifier: GNU GPL v3

"""
Implementations for running StructuralGT via PyPi library
"""

import sys
import logging
from .cli_app import TerminalApp
logger = logging.getLogger("SGT App")
logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s", stream=sys.stdout)

class ExpressGT:
    """Exposes Terminal app to PyPi library."""

    def __init__(self, image_dir: str = "", image_file: str = "", output_dir: str = "", config_file: str = ""):
        """
        Exposes Terminal app methods for executing StructuralGT tasks. Please provide either
        `image_dir` or `image_path`, but not both.

        :param image_dir: Directory contains image files (make sure only the required image files are in this directory).
        :param image_file: Path to the image file.
        :param output_dir: Directory where results files of StructuralGT are stored.
        :param config_file: Path to the configuration file (usually 'sgt_config.ini').
        """
        if (image_dir is None and image_file is None) or (image_dir and image_file):
            raise ValueError("You must provide either `image_path` or `image_dir`, but not both.")
        self._image_dir = image_dir
        self._image_file = image_file
        self._config_file = config_file
        self._output_dir = output_dir

        # 1. Create Terminal App
        self._term_app = TerminalApp(self._config_file)

        # 2. Verify image files
        self._term_app.check_image_files(img_path=self._image_file, img_dir=self._image_dir, out_dir=self._output_dir)

    def process_image(self):
        """Runs StructuralGT task that applies the selected filters on the image."""
        run_multi_gt = True if self._image_dir != "" else False
        if run_multi_gt:
            self._term_app.replicate_sgt_configs()
            keys_list = list(self._term_app.sgt_objs.keys())
            count = len(keys_list)
            for i, key in enumerate(self._term_app.sgt_objs):
                sgt_obj = self._term_app.sgt_objs[key]
                TerminalApp.update_progress(101, f"Processing Image {i+1} of {count}")
                self._term_app.task_worker.task_apply_img_filters(sgt_obj.ntwk_p)
        else:
            sgt_obj = self._term_app.get_selected_sgt_obj()
            self._term_app.task_worker.task_apply_img_filters(sgt_obj.ntwk_p)

    def extract_graph(self):
        """Run StructuralGT task to extract graph."""
        run_multi_gt = True if self._image_dir != "" else False
        if run_multi_gt:
            self._term_app.replicate_sgt_configs()
            keys_list = list(self._term_app.sgt_objs.keys())
            count = len(keys_list)
            for i, key in enumerate(self._term_app.sgt_objs):
                sgt_obj = self._term_app.sgt_objs[key]
                TerminalApp.update_progress(101, f"Extracting Graph {i+1} of {count}")
                self._term_app.task_worker.task_extract_graph(sgt_obj.ntwk_p)
        else:
            sgt_obj = self._term_app.get_selected_sgt_obj()
            self._term_app.task_worker.task_extract_graph(sgt_obj.ntwk_p)

    def compute_gt_descriptors(self):
        """Run StructuralGT task to compute the selected graph theory parameters/descriptors."""
        run_multi_gt = True if self._image_dir != "" else False
        if run_multi_gt:
            self._term_app.replicate_sgt_configs()
            self._term_app.task_worker.task_compute_multi_gt(self._term_app.sgt_objs)
        else:
            self._term_app.task_worker.task_compute_gt(self._term_app.get_selected_sgt_obj())



# SPDX-License-Identifier: GNU GPL v3

"""
QThread Worker class for running tasks in the background.
"""

from PySide6.QtCore import QThread


class QThreadWorker(QThread):
    def __init__(self, func, args, parent=None):
        super().__init__(parent)
        self.func = func  # Store function reference
        self.args = args  # Store arguments

    def run(self):
        if self.func:
            self.func(*self.args)  # Call function with arguments




import cv2
import matplotlib.pyplot as plt
from sgtlib import modules as sgt

# set paths
img_path = "../datasets/InVitroBioFilm.png"
cfg_file = ""   # Optional: leave blank

# Define a function for receiving progress updates
def print_updates(progress_val, progress_msg):
    print(f"{progress_val}: {progress_msg}")

# Create a Network object
ntwk_obj, _ = sgt.ImageProcessor.create_imp_object(img_path, config_file=cfg_file)

# Apply image filters according to cfg_file
ntwk_obj.add_listener(print_updates)
ntwk_obj.apply_img_filters()
ntwk_obj.remove_listener(print_updates)

# View images
bin_images = ntwk_obj.binary_image_3d
mod_images = ntwk_obj.processed_image_3d
plt.imshow(cv2.cvtColor(bin_images[0], cv2.COLOR_BGR2RGB))
plt.axis('off')  # Optional: Turn off axis ticks and labels for a cleaner image display
plt.title('Binary Image')
plt.show()

plt.imshow(mod_images[0], cmap='gray')
plt.axis('off')  # Optional: Turn off axis ticks and labels for a cleaner image display
plt.title('Processed Image')
plt.show()

# Extract graph
ntwk_obj.add_listener(print_updates)
ntwk_obj.build_graph_network()
ntwk_obj.remove_listener(print_updates)

# View graph
net_image = ntwk_obj.graph_image
plt.imshow(net_image)
plt.axis('off')  # Optional: Turn off axis ticks and labels for a cleaner image display
plt.title('Graph Image')
plt.show()

# Compute graph theory metrics
compute_obj = sgt.GraphAnalyzer(ntwk_obj)
sgt.GraphAnalyzer.safe_run_analyzer(compute_obj, print_updates)
print(compute_obj.results_df)

# Save in PDF
sgt.GraphAnalyzer.write_to_pdf(compute_obj)